import sys
import numpy as np

def preprocess_seq(data, seq_length):

    seq_onehot = np.zeros((len(data), 1, seq_length, 4), dtype=float)

    for l in range(len(data)):
        for i in range(seq_length):
            try:
                data[l][i]
            except Exception:
                print(data[l], i, seq_length, len(data))

            if   data[l][i] in "Aa":  seq_onehot[l, 0, i, 0] = 1
            elif data[l][i] in "Cc":  seq_onehot[l, 0, i, 1] = 1
            elif data[l][i] in "Gg":  seq_onehot[l, 0, i, 2] = 1
            elif data[l][i] in "Tt":  seq_onehot[l, 0, i, 3] = 1
            elif data[l][i] in "Xx":  pass
            elif data[l][i] in "Nn.": pass
            else:
                print("[Input Error] Non-ATGC character " + data[l])
                sys.exit()

    return seq_onehot

def seq_concat(data, col1='WT74_On', col2='Edited74_On', seq_length=74):
    wt = preprocess_seq(data[col1], seq_length)
    ed = preprocess_seq(data[col2], seq_length)
    g = np.concatenate((wt, ed), axis=1)
    g = 2 * g - 1

    return g


import requests, platform, inspect, os, sys, subprocess
import pandas as pd
from tqdm import tqdm

from genet import analysis



class CasOFFinder:
    def __init__(self, custom_path:str=None):
        """Cas-OFFinder를 python 함수로 불러와서 사용할 수 있게 하는 wrapper.
        사용하는 환경에 Cas-OFFinder가 설치되어 있어야 한다. 

        Input으로 받는 것은 Cas-OFFinder의 것을 따른다.

        Args:
            custom_path (str): _description_
        """

        self.version = '2.4.1'
        self.path    = inspect.getfile(analysis).replace('__init__.py', '')





        ## 만약 찾아봤는데 없다면

        # Step1: GitHub에서 다운로드 받기
        # 현재 시스템의 운영체제 확인
        os_type = platform.system()

        if os_type == 'Linux':
            self.cas_offinder_path = 'cas-offinder_linux_x86-64'
        elif os_type == 'Darwin':
            self.cas_offinder_path = 'cas-offinder_macos_x86-64'
        elif os_type == 'Windows':
            self.cas_offinder_path = 'cas-offinder_windows_x86-64'
        else:
            raise NotImplementedError('Cas-OFFinder is not supported on this operating system.')
        

        # Step2: zip file을 압축 해제하기
        import zipfile
    	
        fantasy_zip = zipfile.ZipFile('cas-offinder_linux_x86-64.zip')
        fantasy_zip.extractall('./')
        fantasy_zip.close()

        # Step3: 다시 cas-offinder 경로를 지정해서 불러오기

        
        

        pass

    def run(self, 
            input:list, output_path:str, 
            n_mismatch:int=3, n_dna_bulge:int=0, n_rna_bulge:int=0, 
            pam:str='NGG', use_gpu=True) -> pd.DataFrame:
        '''Cas-OFFinder를 돌려서 output file을 만들어주는 함수.
        Cas-OFFinder의 input format을 맞출 필요 없이, 내부적으로 tsv 파일을 만들어줌.'''


        df_out = pd.read_csv()

        
        return df_out


    def download_github_release(self, repo:str, release_tag:str, file_name:str, save_dir:str):
        """A function to download files from a specified path in a GitHub repository. 
        It uses `requests.get` to download from the GitHub URL.

        Args:
            repo (str): The GitHub repository where the file to be downloaded is located.
            release_tag (str): The path within the repository where the file is located.
            file_name (str): The name of the file to download.
            save_dir (str): The location to download the file.
        """        
        
        print(f'The {repo} {release_tag} is not installed. Download file from GitHub.')

        url = f"https://github.com/{repo}/releases/download/{release_tag}/{file_name}"
        save_path = f"{save_dir}/{file_name}"

        response = requests.get(url, stream=True)
        response.raise_for_status()
        total_size = int(response.headers.get('content-length', 0))
        
        with open(save_path, "wb") as file:
            
            for data in tqdm(response.iter_content(chunk_size=4096), total=total_size//4096, 
                                unit="KB", desc="Downloading", unit_scale=True, leave=False):
                
                file.write(data)
        
        print(f"[Info] File downloaded: {save_path}")

    # def End: download_github_release




import os, sys
import genet.utils
from Bio import Entrez, SeqIO

'''
Branch test 221230
'''

class GetGene:
    '''
    NCBI에서 reference gene을 찾기 위한 함수.
    기본적으로 biopython의 Entrez module과 GenBank module을 사용한다. 

    default는 human genome에서 reference gene sequence를 가져오는 것이다. \n
    하지만 만약 사용자가 다른 sequence 정보를 가져오고 싶다면,\n
    search_option에서 원하는 조건식을 넣어주면 되고,
    species도 다른 것으로 바꿔줄 수 있다.  

    하지만 species를 적어줄 때에는 정확한 학명으로 적어줘야 한다.\n
    Human (X) / Homo sapiens (O) > 확인 필요, human도 될 듯?\n
    Mouse (X) / Mus musculus (O)

    만약 Mus musculus라고 전체로 적어주지 않고 Mus라고만 적으면,\n
    정확한 RefSeq이 찾아지지 않는다. \n
    학명에 Mus를 포함하는 것이 총 3종류가 있기 때문이다.\n
    Mus caroli / Mus musculus / Mus pahari
    
    ## example:
    >>> from genet import database as db

    >>> # To get BRCA1 gene sequence information from mouse
    >>> gene = db.GetGene('BRCA1', species='Mus musculus')

    '''
    def __init__(self, 
                 gene_name:str,
                 species:str = 'Homo sapiens',
                 search_option:str = 'AND biomol_genomic[PROP] AND RefSeqGene[Filter]',
                 ):

        print('Please enter your email')

        email = input('Please enter your email address to access NCBI database: ')
        Entrez.email = email


        print('Find %s from NCBI nucleotide database' % gene_name)
        search_string = '%s[Gene] AND %s[title] AND %s[Organism] %s' % (gene_name, gene_name, species, search_option)
        
        self.handle      = Entrez.esearch(db="nucleotide", term=search_string)
        self.gene_record = Entrez.read(self.handle)
        self.ids         = self.gene_record['IdList']
        if len(self.gene_record['IdList']) > 1:
            print('[Warnning] There are more than one ID from result. Please check your search options.')


        print('RefGenID found: ', self.ids)
        print('')

        self.fetch      = Entrez.efetch(db='nucleotide', id=self.gene_record['IdList'], rettype='gb', retmode='xlm')
        self.seq_record = SeqIO.read(self.fetch, 'genbank')

    
    # def __init__: End

    def is_misc_feat(self, feat): return feat.type == 'misc_feature'
    def is_source(self, feat):    return feat.type == 'source'

    def exons(self):
        '''
        esearch로 가져온 RefSeq의 ID를 받아서, efetch로 정보를 불러온다.
        불러온 정보는 seq_record로 저장되고, 그 안에서 각종 정보를 가져올 수 있다.
        
        '''
        def is_exon(feat):      return feat.type == 'exon'

        self.feat = self.seq_record.features

        list_exons = [f for f in filter(is_exon, self.feat)]
        return list_exons

    def transcripts(self):
        '''
        esearch로 가져온 RefSeq의 ID를 받아서, efetch로 정보를 불러온다.
        불러온 정보는 seq_record로 저장되고, 그 안에서 각종 정보를 가져올 수 있다.
        
        '''
        def is_mrna(feat):      return feat.type == 'mRNA'

        self.feat = self.seq_record.features
        list_transcripts = [f for f in filter(is_mrna, self.feat)]

        return list_transcripts

    def cds(self):
        '''
        esearch로 가져온 RefSeq의 ID를 받아서, efetch로 정보를 불러온다.
        불러온 정보는 seq_record로 저장되고, 그 안에서 각종 정보를 가져올 수 있다.
        
        '''
        def is_cds(feat):      return feat.type == 'CDS'

        self.feat = self.seq_record.features
        list_transcripts = [f for f in filter(is_cds, self.feat)]

        return list_transcripts
    

    def misc(self):
        '''
        esearch로 가져온 RefSeq의 ID를 받아서, efetch로 정보를 불러온다.
        불러온 정보는 seq_record로 저장되고, 그 안에서 각종 정보를 가져올 수 있다.
        
        '''
        def is_misc_feat(feat): return feat.type == 'misc_feature'

        self.feat = self.seq_record.features
        list_transcripts = [f for f in filter(is_misc_feat, self.feat)]

        return list_transcripts

    def source(self):
        '''
        esearch로 가져온 RefSeq의 ID를 받아서, efetch로 정보를 불러온다.
        불러온 정보는 seq_record로 저장되고, 그 안에서 각종 정보를 가져올 수 있다.
        
        '''
        def is_source(feat):    return feat.type == 'source'

        self.feat = self.seq_record.features
        list_transcripts = [f for f in filter(is_source, self.feat)]

        return list_transcripts
    

class GetClinVar:
    '''
    NCBI ClinVar에서 record를 찾기위한 function.\n
    기본적으로 biopython의 Entrez module을 사용한다. 

    ## example:
    >>> from genet import database as db
    >>> cv_record = db.GetClinVar('VCV000428864')

    GetClinVar class에서 seq method로 해당 record sequence를 fetching 할 수 있다.

    ## example:
    >>> ref_seq, alt_seq = cv_record.seq()

    output:
    ref_seq = 'GGTCACTCACCTGGAGTGAGCCCTGCTCCCCCCTGGCTCCTTCCCAGCCTGGGCATCCTTGAGTTCCAAGGCCTCATTCAGCTCTCGGAACATCTCGAAGCGCTCACGCCCACGGATCTGC'
    alt_seq = 'GGTCACTCACCTGGAGTGAGCCCTGCTCCCCCCTGGCTCCTTCCCAGCCTGGGCATCCTTGTTCCAAGGCCTCATTCAGCTCTCGGAACATCTCGAAGCGCTCACGCCCACGGATCTGCAG'


    seq method에서 int 값을 넣어주면, context 길이를 조절할 수 있다.
    >>> ref_seq, alt_seq = cv_record.seq(80)

    '''

    def __init__(self, record_id:str):

        email = input('Please enter your email address to access NCBI database: ')
        Entrez.email = email


        self._record_id = record_id

        if self._record_id.startswith('VCV'):
            self.handle = Entrez.efetch(db='clinvar', id=self._record_id.split('.')[0], rettype='vcv') # VCV로 받을 경우    
        else:            
            self.handle = Entrez.efetch(db='clinvar', id=self._record_id, rettype='vcv', is_varationid='true', from_esearch="true") # variation ID로 받을 경우
        
        import xml.etree.ElementTree as ET
        self.result = ET.parse(self.handle)
        self.root = self.result.getroot()
        
        self.var_loc = self.root.findall('./VariationArchive/InterpretedRecord/SimpleAllele/Location/SequenceLocation')

        for self.info in self.var_loc:
            print(self.info)
            if self.info.attrib['Assembly'] == 'GRCh38':
                self.chr_acc = self.info.attrib['Accession']
                self.start   = int(self.info.attrib['start'])
                self.stop    = int(self.info.attrib['stop'])
                self.ref_nt  = self.info.attrib['referenceAlleleVCF']
                self.alt_nt  = self.info.attrib['alternateAlleleVCF']
                self.alt_len = int(self.info.attrib['variantLength'])
                break

        if   len(self.ref_nt) == len(self.alt_nt): self.alt_type = 'sub'
        elif len(self.ref_nt) <  len(self.alt_nt): self.alt_type = 'ins'
        elif len(self.ref_nt) >  len(self.alt_nt): self.alt_type = 'del'
    
    # def __init__: End

    def seq(self, context:int = 60):
        '''
        esearch로 가져온 RefSeq의 ID를 받아서, efetch로 정보를 불러온다.
        불러온 정보는 seq_record로 저장되고, 그 안에서 각종 정보를 가져올 수 있다.
        
        '''
        self.chr_seq_fetch = Entrez.efetch(db="nucleotide", 
                                           id=self.chr_acc, 
                                           rettype="fasta", 
                                           strand=1, 
                                           seq_start = self.start-context, 
                                           seq_stop  = self.stop+context+self.alt_len
                                           )

        self.ref_seq = str(SeqIO.read(self.chr_seq_fetch, "fasta").seq)
        self.chr_seq_fetch.close()
        
        if self.alt_type != 'del':
            self.alt_seq = self.ref_seq[:context] + self.alt_nt + self.ref_seq[context+1:]
        else:
            self.alt_seq = self.ref_seq[:context] + self.ref_seq[context+self.alt_len:]

        if self.alt_type == 'ins':
            self.ref_seq = self.ref_seq[1:]
            self.alt_seq = self.alt_seq[1:]

        return self.ref_seq[:1+context*2], self.alt_seq[:1+context*2]





from genet.predict.functional import *
from genet.predict.Nuclease import *
from genet.predict.PrimeEditor import *

#### Under development ###
# from genet.predict.DeepSmallCas9 import *
# from genet.predict.functional_dev import *

#### Will be removed ###
# from genet.predict.DeepCas9Variants import *

from silence_tensorflow import silence_tensorflow
import warnings

silence_tensorflow()
warnings.filterwarnings(action='ignore')



# genet package and modules 
from genet.predict.PredUtils import *
from genet.predict.Nuclease import SpCas9
from genet.models import LoadModel

# python standard packages
import os, sys, regex, gzip
import ray
import numpy as np
import pandas as pd
from glob import glob
from tqdm import tqdm
from time import time

# pytorch package and modules 
import torch
import torch.nn.functional as F
import torch.nn as nn
from torch.utils.data import TensorDataset, DataLoader

# biopython package and modules 
from Bio import SeqIO
from Bio.Seq import Seq, transcribe, back_transcribe, reverse_complement, reverse_complement_rna
from Bio.SeqUtils import MeltingTemp as mt
from Bio.SeqUtils import gc_fraction as gc

from RNA import fold_compound

np.set_printoptions(threshold=sys.maxsize)
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'


class DeepPrime:

    def __init__(self, sequence: str, name:str='SampleName', pam:str = 'NGG',
                 pbs_min:int = 7, pbs_max:int = 15,
                 rtt_min:int = 0, rtt_max:int = 40, 
                 spacer_len:int=20,
                ):
        """DeepPrime: pegRNA activity prediction models

        Args:
            sequence (str): Sequence to prime editing. Intended prime editing should marked with parenthesis.
            name (str, optional): Sample ID for pegRNAs. Defaults to 'SampleName'
            pam (str, optional): PAM sequence. Available PAMs are NGG, NGA, NAG, NRCH. Defaults to 'NGG'.
            pbs_min (int, optional): Minimum length of PBS (1-17). Defaults to 7.
            pbs_max (int, optional): Maximum length of PBS (1-17). Defaults to 15.
            rtt_min (int, optional): Minimum length of RTT (0-40). Defaults to 0.
            rtt_max (int, optional): Maximum length of RTT (0-40). Defaults to 40.
        """        
        
        # input parameters
        self.input_params = {
            'name'      : name,
            'sequence'  : sequence,
            'nAltIndex' : 60,
            'spacerlen' : spacer_len,
            'pam'       : pam,
            'pbs_min'   : pbs_min,
            'pbs_max'   : pbs_max,
            'rtt_min'   : rtt_min,
            'rtt_max'   : rtt_max,
            'wt_seq'    : None,
            'ed_seq'    : None,
            'edit_type' : None,
            'edit_len'  : None,
        }

        # Check input parameters
        self.dp_input = DeepPrimeInputProcessor(self.input_params)
        self.input_params = self.dp_input.check_input()

        ## DeepPrime Feature Extraction 
        cFeat = PEFeatureExtraction(self.input_params)

        cFeat.input_id = name
        cFeat.get_all_RT_PBS(self.input_params)
        cFeat.make_rt_pbs_combinations()
        cFeat.determine_seqs()
        cFeat.determine_secondary_structure()
        self.features = cFeat.make_output_df()
        
        del cFeat

        if len(self.features) > 0:
            self.list_Guide30 = [WT74[:30] for WT74 in self.features['Target']]
            self.features['DeepSpCas9_score'] = SpCas9().predict(self.list_Guide30)['SpCas9']
            self.pegRNAcnt = len(self.features)
        
        else:
            print('\nsID:', name)
            print('DeepPrime only support RTT length upto 40nt')
            print('There are no available pegRNAs, please check your input sequences.\n')
            self.pegRNAcnt = 0

    # def __init__: END


    def predict(self, pe_system:str, cell_type:str = 'HEK293T', show_features:bool = False, report=False) -> pd.DataFrame:
        """_summary_
    
        Args:
            pe_system (str): Available PE systems are PE2, PE2max, PE4max, NRCH_PE2, NRCH_PE2max, NRCH_PE4max
            cell_type (str, optional): Available Cell types are HEK293T, HCT116, MDA-MB-231, HeLa, DLD1, A549, NIH3T3. Defaults to 'HEK293T'.
            show_features (bool, optional): _description_. Defaults to False.
            report (bool, optional): _description_. Defaults to False.

        Returns:
            pd.DataFrame: 각 pegRNA와 target쌍 마다의 DeepPrime prediction score를 계산한 결과를 DataFrame으로 반환.
        """
        
        # Load models
        model_info = LoadModel('DeepPrime', pe_system, cell_type)
        model_dir  = model_info.model_dir

        # Check pe_system is available for PAM
        self.dp_input.check_pe_type(pe_system)

        # Data preprocessing for deep learning model
        df_all = self.features.copy()

        os.environ['CUDA_VISIBLE_DEVICES']='0'
        device = 'cuda' if torch.cuda.is_available() else 'cpu'

        mean = pd.read_csv(f'{model_dir}/mean_231124.csv', header=None, index_col=0).squeeze()
        std  = pd.read_csv(f'{model_dir}/std_231124.csv',  header=None, index_col=0).squeeze()

        test_features = select_cols(df_all)

        g_test = seq_concat(df_all)
        x_test = (test_features - mean) / std

        g_test = torch.tensor(g_test, dtype=torch.float32, device=device)
        x_test = torch.tensor(x_test.to_numpy(), dtype=torch.float32, device=device)

        models = [m_files for m_files in glob(f'{model_dir}/*.pt')]
        preds  = []

        for m in models:
            model = GeneInteractionModel(hidden_size=128, num_layers=1).to(device)
            model.load_state_dict(torch.load(m, map_location=device))
            model.eval()
            with torch.no_grad():
                g, x = g_test, x_test
                g = g.permute((0, 3, 1, 2))
                pred = model(g, x).detach().cpu().numpy()
            preds.append(pred)
        
        # AVERAGE PREDICTIONS
        preds = np.squeeze(np.array(preds))
        preds = np.mean(preds, axis=0)
        preds = np.exp(preds) - 1

        df_all.insert(1, f'{pe_system}_score', preds)


        if   show_features == False: return df_all.iloc[:, :11]
        elif show_features == True : return df_all

    # def predict: END


class DeepPrimeInputProcessor:
    def __init__(self, input_params:dict):
        """Make DeepPrime inputs from input sequence.
        
        Args:
            sequence (str): Sequence to prime editing. Intended prime editing should marked with parenthesis.

        Raises:
            ValueError: _description_

        Returns:
            dict: DeepPrime inputs from input sequence
        """    

        self.input_params = input_params

        sequence = self.input_params['sequence']

        f_context, res_seq  = sequence.split('(')
        edit_seq, r_context = res_seq.split(')')
        wt, ed = edit_seq.split('/')

        if len(wt) == len(ed): edit_type = 'sub'; edit_len = len(ed)
        elif wt == '': edit_type = 'ins'; edit_len = len(ed)
        elif ed == '': edit_type = 'del'; edit_len = len(wt)
        else: raise ValueError('Not supported edit type.')

        self.input_params.update({
            'wt_seq': f_context[-60:] + (wt + r_context)[:61],
            'ed_seq' : f_context[-60:] + (ed + r_context)[:61],
            'edit_type': edit_type,
            'edit_len': edit_len,
        })
        

    def check_input(self):
        
        if len(self.input_params['wt_seq']) != 121:
            raise ValueError('Please check your input sequence. The length of context sequence is not enough.')
        
        if len(self.input_params['ed_seq']) != 121:
            raise ValueError('Please check your input sequence. The length of context sequence is not enough.')

        if self.input_params['pbs_min'] < 1:
            raise ValueError('Please check your input: pbs_min. Please set PBS max length at least 1nt')
        
        if self.input_params['pbs_max'] > 17:
            raise ValueError('Please check your input: pbs_max. Please set PBS max length upto 17nt')
        
        if self.input_params['rtt_max'] > 40:
            raise ValueError('Please check your input: rtt_max. Please set RTT max length upto 40nt')

        if self.input_params['edit_type'] not in ['sub', 'ins', 'del']:
            raise ValueError('Please check your input: edit_type. Available edit style: sub, ins, del')
        
        if self.input_params['pam'] not in ['NGG', 'NRCH', 'NAG', 'NGA', 'NNGG']:
            raise ValueError('Please check your input: pam. Available PAM: NGG, NGA, NAG, NRCH, NNGG')

        if self.input_params['edit_len'] > 3:
            raise ValueError('Please check your input: edit_len. Please set edit length upto 3nt. Available edit length range: 1~3nt')
        
        if self.input_params['edit_len'] < 1:
            raise ValueError('Please check your input: edit_len. Please set edit length at least 1nt. Available edit length range: 1~3nt')

        return self.input_params
    
    # def check_input: END

    def check_pe_type(self, pe_system:str):

        pam = self.input_params['pam']
        
        if pam == 'NRCH' and pe_system not in ['NRCH_PE2', 'NRCH_PE2max', 'NRCH_PE4max']:
            raise ValueError(f'{pam} PAM is not available for {pe_system}. Please check PAM or pe_system. ')
        
        if pam == 'NNGG' and pe_system not in ['sRGN-PE2max']:
            raise ValueError(f'{pam} PAM is not available for {pe_system}. Please check PAM or pe_system. ')

        return None
    
    # def check_input: END


class DeepPrimeBatch:
    def __init__(self, data, pam:str = 'NGG',
                 pbs_min:int = 7, pbs_max:int = 15,
                 rtt_min:int = 0, rtt_max:int = 40, 
                 spacer_len:int=20,
                ):
        """DeepPrime을 Batch Mode로 돌릴 수 있는 pipeline. 
        FeatureExtraction을 CPU에서 multiprocessing을 전체적으로 하고, 
        한번에 많은 양의 dataset을 DataLoader로 밀어넣어서 GPU 연산을 하도록 설계

        Args:
            sequence (str): Sequence to prime editing. Intended prime editing should marked with parenthesis.
            name (str, optional): Sample ID for pegRNAs. Defaults to 'SampleName'
            pam (str, optional): PAM sequence. Available PAMs are NGG, NGA, NAG, NRCH. Defaults to 'NGG'.
            pbs_min (int, optional): Minimum length of PBS (1-17). Defaults to 7.
            pbs_max (int, optional): Maximum length of PBS (1-17). Defaults to 15.
            rtt_min (int, optional): Minimum length of RTT (0-40). Defaults to 0.
            rtt_max (int, optional): Maximum length of RTT (0-40). Defaults to 40.
        """        
        
        # input parameters
        try:
            self.df_input = pd.DataFrame(data)
        except:
            self.df_input = self.load_file_as_dataframe(data)


        self.input_params = {
            'nAltIndex' : 60,
            'spacerlen' : spacer_len,
            'pam'       : pam,
            'pbs_min'   : pbs_min,
            'pbs_max'   : pbs_max,
            'rtt_min'   : rtt_min,
            'rtt_max'   : rtt_max,
            'wt_seq'    : None,
            'ed_seq'    : None,
            'edit_type' : None,
            'edit_len'  : None,
        }



    def preprocess(self, num_cpus=1, memory=2) -> pd.DataFrame:

        ray.init()

       # 병렬 작업을 위한 Ray Actor 클래스 정의
        @ray.remote(num_cpus=num_cpus, memory=memory*1024*1024*1024)
        class DeepPrimeWorker:
            def __init__(self, sequence:str, id:str, input_params:dict):

                self.pegrna = DeepPrime(sequence, name=id)
            
            def get_feature(self):
                return self.pegrna.features

        # 데이터프레임에서 각 시퀀스를 병렬로 처리
        self.workers = [DeepPrimeWorker.remote(
            data['sequence'], data['id'], self.input_params
            ) for _, data in self.df_input.iterrows()]
        
        # Ray 작업 실행 및 결과 수집
        self.features = ray.get([worker.get_feature.remote() for worker in self.workers])

    # def __init__: END


    def predict(self, pe_system:str, cell_type:str = 'HEK293T', show_features:bool = False, gpu_id='0', report=False) -> pd.DataFrame:
        """_summary_
    
        Args:
            pe_system (str): Available PE systems are PE2, PE2max, PE4max, NRCH_PE2, NRCH_PE2max, NRCH_PE4max
            cell_type (str, optional): Available Cell types are HEK293T, HCT116, MDA-MB-231, HeLa, DLD1, A549, NIH3T3. Defaults to 'HEK293T'.
            show_features (bool, optional): _description_. Defaults to False.
            report (bool, optional): _description_. Defaults to False.

        Returns:
            pd.DataFrame: 각 pegRNA와 target쌍 마다의 DeepPrime prediction score를 계산한 결과를 DataFrame으로 반환.
        """
        

        # Load models
        model_info = LoadModel('DeepPrime', pe_system, cell_type)
        model_dir  = model_info.model_dir

        # Check pe_system is available for PAM
        # 이 부분은 구현해야 함. 

        # Data preprocessing for deep learning model
        df_all = self.features.copy()

        os.environ['CUDA_VISIBLE_DEVICES'] = gpu_id
        device = 'cuda' if torch.cuda.is_available() else 'cpu'

        mean = pd.read_csv(f'{model_dir}/mean_231124.csv', header=None, index_col=0).squeeze()
        std  = pd.read_csv(f'{model_dir}/std_231124.csv',  header=None, index_col=0).squeeze()

        test_features = select_cols(df_all)

        g_test = seq_concat(df_all)
        x_test = (test_features - mean) / std

        # VRAM 절약을 위해 데이터셋을 CPU에 남겨두고 Loading할 배치만 GPU로 이동
        g_test_tensor = torch.tensor(g_test, dtype=torch.float32)  # CPU Tensor
        x_test_tensor = torch.tensor(x_test.to_numpy(), dtype=torch.float32)  # CPU Tensor

        # g_test = torch.tensor(g_test, dtype=torch.float32, device=device)
        # x_test = torch.tensor(x_test.to_numpy(), dtype=torch.float32, device=device)

        # DataLoader 정의
        batch_size = 4096
        num_workers = 4 # could be modified; default: 0; dataloader에서 vram으로 올려주는 역할인데, worker 개수만큼 쓰레드를 만들어서 대기하고 있는 것

        test_set = TensorDataset(g_test_tensor, x_test_tensor)
        test_loader = DataLoader(
            dataset=test_set,
            batch_size=batch_size,
            shuffle=False, # 학습 때에는 shuffle 하는 것이 좋지만, 실제 사용 할 때에는 순서대로 하기 위함
            num_workers=num_workers,
            pin_memory=True # for faster VRAM loading CPU 안에 있는 메모리 
        )

        models = [m_files for m_files in glob(f'{model_dir}/*.pt')]
        
        total_size = len(test_loader.dataset)  # 전체 데이터의 크기
        all_preds = np.zeros((len(models), total_size, 1), dtype=np.float32)

        for model_idx, m in enumerate(models):
            model = GeneInteractionModel(hidden_size=128, num_layers=1).to(device)
            model.load_state_dict(torch.load(m, map_location=device))
            model = nn.DataParallel(model)  # DataParallel을 사용하여 여러 GPU에 모델 분산
            model.eval()

            current_idx = 0

            with torch.no_grad():
                for g, x in test_loader:
                    
                    # non_blocking: loading 순서에 영향 있는지?
                    # async loading이 만약 충돌을 일으켜서 error가 나오면 끄는 것으로 하기
                    g = g.permute((0, 3, 1, 2)).to(device, non_blocking=True)
                    x = x.to(device, non_blocking=True)

                    # 모델 예측 수행
                    pred = model(g, x).detach().cpu().numpy()

                    batch_size = pred.shape[0]
                    all_preds[model_idx, current_idx:current_idx + batch_size] = pred
                    current_idx += batch_size

        # AVERAGE PREDICTIONS
        preds = np.mean(all_preds, axis=0)
        preds = np.exp(preds) - 1

        df_all.insert(1, f'{pe_system}_score', preds)

        if   show_features == False: return df_all.iloc[:, :11]
        elif show_features == True : return df_all

    # def predict: END

    def load_file_as_dataframe(self, file_path):
        # 파일 확장자 추출
        file_extension = file_path.split('.')[-1].lower()
        
        # 확장자에 따라 파일 읽기
        if file_extension == 'csv':
            df = pd.read_csv(file_path)
        elif file_extension == 'txt':
            df = pd.read_csv(file_path, delimiter='\t')
        elif file_extension == 'parquet':
            df = pd.read_parquet(file_path)
        elif file_extension == 'feather':
            df = pd.read_feather(file_path)
        else:
            raise ValueError(f"Unsupported file format: {file_extension}")
        
        df.columns = ['id', 'sequence']
        
        return df



def set_alt_position_window(sStrand, sAltKey, nAltIndex, nIndexStart, nIndexEnd, nAltLen):
    if sStrand == '+':

        if sAltKey.startswith('sub'):
            return (nAltIndex + 1) - (nIndexStart - 3)
        else:
            return (nAltIndex + 1) - (nIndexStart - 3)

    else:
        if sAltKey.startswith('sub'):
            return nIndexEnd - nAltIndex + 3 - (nAltLen - 1)

        elif sAltKey.startswith('del'):
            return nIndexEnd - nAltIndex + 3 - nAltLen

        else:
            return nIndexEnd - nAltIndex + 3 + nAltLen
        # if END:
    # if END:

# def END: set_alt_position_window


def set_PAM_nicking_pos(sStrand, sAltType, nAltLen, nAltIndex, nIndexStart, nIndexEnd):
    if sStrand == '-':
        nPAM_Nick = nIndexEnd + 3
    else:
        nPAM_Nick = nIndexStart - 3

    return nPAM_Nick

# def END: set_PAM_Nicking_Pos


def check_PAM_window(dict_sWinSize, sStrand, nIndexStart, nIndexEnd, sAltType, nAltLen, nAltIndex):
    nUp, nDown = dict_sWinSize[sAltType][nAltLen]

    if sStrand == '+':
        nPAMCheck_min = nAltIndex - nUp + 1
        nPAMCheck_max = nAltIndex + nDown + 1
    else:
        # nPAMCheck_min = nAltIndex - nDown + 1
        nPAMCheck_min = nAltIndex - nDown
        nPAMCheck_max = nAltIndex + nUp + 1
    # if END:

    if nIndexStart < nPAMCheck_min or nIndexEnd > nPAMCheck_max:
        return 0
    else:
        return 1

# def END: check_PAM_window

class PEFeatureExtraction:
    def __init__(self, input_params: dict):
        
        # configuration parameters
        # 이게 sRGN을 지정하는 단계까 .predict  부분이기 때문에, feature extraction 부분에서 spacer 길이를 조정할 수가 없음...!
        # pe_system을 지정하는 단계를 바꿔줘야 할 것 같음. 
        self.spacer_len = input_params['spacerlen']
        
        # Initialized
        self.get_input(input_params)
        self.get_sAltNotation(input_params['nAltIndex'])

        self.dict_sSeqs   = {}
        self.dict_sCombos = {}
        self.dict_sOutput = {}

        self.gc_cache   = {}
        self.tm_cache   = {}
        self.mfe_cache  = {}

    # def End: __init__
    
    def get_input(self, input_params: dict):
        self.sWTSeq     = input_params['wt_seq'].upper()
        self.sEditedSeq = input_params['ed_seq'].upper()
        self.sAltType   = input_params['edit_type']
        self.nAltLen    = input_params['edit_len']
        self.sAltKey    = self.sAltType + str(self.nAltLen)

        if   self.sAltType.startswith('sub'): self.type_sub = 1; self.type_del = 0; self.type_ins = 0
        elif self.sAltType.startswith('del'): self.type_del = 1; self.type_sub = 0; self.type_ins = 0
        elif self.sAltType.startswith('ins'): self.type_ins = 1; self.type_del = 0; self.type_sub = 0
    
    # def End: get_input

    def get_sAltNotation(self, nAltIndex):
        if self.sAltType == 'sub':
            self.sAltNotation = '%s>%s' % (
                self.sWTSeq[nAltIndex:nAltIndex + self.nAltLen], self.sEditedSeq[nAltIndex:nAltIndex + self.nAltLen])

        elif self.sAltType == 'del':
            self.sAltNotation = '%s>%s' % (
                self.sWTSeq[nAltIndex:nAltIndex + 1 + self.nAltLen], self.sEditedSeq[nAltIndex])

        else:
            self.sAltNotation = '%s>%s' % (
                self.sWTSeq[nAltIndex], self.sEditedSeq[nAltIndex:nAltIndex + self.nAltLen + 1])

    # def END: get_sAltNotation

    def get_all_RT_PBS(self, input_params:dict):
        """
        nMinPBS: If you set specific number, lower than MinPBS will be not generated. Default=0
        nMaxPBS: If you set specific number, higher than MinPBS will be not generated. Default=17
        nMaxRT = : If you set specific number, higher than MinPBS will be not generated. Default=40
        nSetPBSLen = 0  # Fix PBS Len: Set if >0
        nSetRTLen = 0  # Fix RT  Len: Set if >0
        PAM: 4-nt sequence
        """

        nAltIndex   = input_params['nAltIndex']
        pam         = input_params['pam']
        nMinPBS     = input_params['pbs_min']
        nMaxPBS     = input_params['pbs_max']
        nMaxRT      = input_params['rtt_max']
        nSetPBSLen  = 0
        nSetRTLen   = 0

        nMaxEditPosWin = nMaxRT + 3  # Distance between PAM and mutation

        dict_sWinSize = {'sub': {1: [nMaxRT - 1 - 3, 6], 2: [nMaxRT - 2 - 3, 6], 3: [nMaxRT - 3 - 3, 6]},
                        'ins': {1: [nMaxRT - 2 - 3, 6], 2: [nMaxRT - 3 - 3, 6], 3: [nMaxRT - 4 - 3, 6]},
                        'del': {1: [nMaxRT - 1 - 3, 6], 2: [nMaxRT - 1 - 3, 6], 3: [nMaxRT - 1 - 3, 6]}}

        
        if pam == 'NRCH': # for NRCH-PE PAM
            dict_sRE = {'+': '[ACGT][ACGT]G[ACGT]|[ACGT][CG]A[ACGT]|[ACGT][AG]CC|[ATCG]ATG', 
                        '-': '[ACGT]C[ACGT][ACGT]|[ACGT]T[CG][ACGT]|G[GT]T[ACGT]|ATT[ACGT]|CAT[ACGT]|GGC[ACGT]|GTA[ACGT]'} 
        elif pam == 'NGG':
            dict_sRE = {'+': '[ACGT]GG[ACGT]', '-': '[ACGT]CC[ACGT]'} # for Original-PE PAM
        elif pam == 'NAG':
            dict_sRE = {'+': '[ACGT]AG[ACGT]', '-': '[ACGT]CT[ACGT]'} # for Original-PE PAM
        elif pam == 'NGA':
            dict_sRE = {'+': '[ACGT]GA[ACGT]', '-': '[ACGT]TC[ACGT]'} # for Original-PE PAM
        elif pam == 'NNGG':
            dict_sRE = {'+': '[ACGT][ACGT]GG', '-': 'CC[ACGT][ACGT]'} # for sRGN-PE PAM

        for sStrand in ['+', '-']:

            sRE = dict_sRE[sStrand]
            for sReIndex in regex.finditer(sRE, self.sWTSeq, overlapped=True):

                if sStrand == '+':
                    nIndexStart = sReIndex.start()
                    nIndexEnd = sReIndex.end() - 1
                    sPAMSeq = self.sWTSeq[nIndexStart:nIndexEnd]
                    sGuideSeq = self.sWTSeq[nIndexStart - self.spacer_len:nIndexEnd]
                else:
                    nIndexStart = sReIndex.start() + 1
                    nIndexEnd = sReIndex.end()
                    sPAMSeq = reverse_complement(self.sWTSeq[nIndexStart:nIndexEnd])
                    sGuideSeq = reverse_complement(self.sWTSeq[nIndexStart:nIndexEnd + 20])

                nAltPosWin = set_alt_position_window(sStrand, self.sAltKey, nAltIndex, nIndexStart, nIndexEnd,
                                                    self.nAltLen)

                ## AltPosWin Filter ##
                if nAltPosWin <= 0:             continue
                if nAltPosWin > nMaxEditPosWin: continue

                nPAM_Nick = set_PAM_nicking_pos(sStrand, self.sAltType, self.nAltLen, nAltIndex, nIndexStart, nIndexEnd)

                if not check_PAM_window(dict_sWinSize, sStrand, nIndexStart, nIndexEnd, self.sAltType, self.nAltLen,
                                        nAltIndex): continue

                sPAMKey = '%s,%s,%s,%s,%s,%s,%s' % (
                    self.sAltKey, self.sAltNotation, sStrand, nPAM_Nick, nAltPosWin, sPAMSeq, sGuideSeq)

                dict_sRT, dict_sPBS = self.determine_PBS_RT_seq(sStrand, nMinPBS, nMaxPBS, nMaxRT, nSetPBSLen,
                                                        nSetRTLen, nAltIndex, nPAM_Nick, nAltPosWin, self.sEditedSeq)

                nCnt1, nCnt2 = len(dict_sRT), len(dict_sPBS)
                if nCnt1 == 0: continue
                if nCnt2 == 0: continue
                
                if sPAMKey not in self.dict_sSeqs:
                    self.dict_sSeqs[sPAMKey] = ''
                self.dict_sSeqs[sPAMKey] = [dict_sRT, dict_sPBS]

            # loop END: sReIndex
        # loop END: sStrand


    # def END: get_all_RT_PBS

    
    def determine_PBS_RT_seq(self, sStrand, nMinPBS, nMaxPBS, nMaxRT, nSetPBSLen, nSetRTLen, nAltIndex, nPAM_Nick,
                            nAltPosWin, sForTempSeq):
        dict_sPBS = {}
        dict_sRT = {}

        list_nPBSLen = [nNo + 1 for nNo in range(nMinPBS, nMaxPBS)]
        for nPBSLen in list_nPBSLen:

            ## Set PBS Length ##
            if nSetPBSLen:
                if nPBSLen != nSetPBSLen: continue

            if sStrand == '+':
                nPBSStart = nPAM_Nick - nPBSLen  # 5' -> PamNick
                nPBSEnd = nPAM_Nick
                sPBSSeq = sForTempSeq[nPBSStart:nPBSEnd] # sForTempSeq = self.EditedSeq

            else:
                if self.sAltKey.startswith('sub'):
                    nPBSStart = nPAM_Nick
                elif self.sAltKey.startswith('ins'):
                    nPBSStart = nPAM_Nick + self.nAltLen
                elif self.sAltKey.startswith('del'):
                    nPBSStart = nPAM_Nick - self.nAltLen

                sPBSSeq = reverse_complement(sForTempSeq[nPBSStart:nPBSStart + nPBSLen]) # sForTempSeq = self.EditedSeq

            # if END: sStrand

            sKey = len(sPBSSeq)
            if sKey not in dict_sPBS:
                dict_sPBS[sKey] = ''
            dict_sPBS[sKey] = sPBSSeq
        # loop END: nPBSLen

        if sStrand == '+':
            if self.sAltKey.startswith('sub'):
                list_nRTPos = [nNo + 1 for nNo in range(nAltIndex + self.nAltLen, (nPAM_Nick + nMaxRT))] # OK
            elif self.sAltKey.startswith('ins'):
                list_nRTPos = [nNo + 1 for nNo in range(nAltIndex + self.nAltLen, (nPAM_Nick + nMaxRT))] # OK
            else:
                list_nRTPos = [nNo + 1 for nNo in range(nAltIndex, (nPAM_Nick + nMaxRT))] ## 수정! ## del2 RHA 3 del1 RHA2
        else:
            if self.sAltKey.startswith('sub'):
                list_nRTPos = [nNo for nNo in range(nPAM_Nick - 1 - nMaxRT, nAltIndex)] ## 수정! ## sub1 sub 3 RHA 0
            else:
                list_nRTPos = [nNo for nNo in range(nPAM_Nick - 3 - nMaxRT, nAltIndex + self.nAltLen - 1)] ## 수정! ## ins2 최소가 2까지 ins3 RHA 최소 3 #del2 RHA 2 del1 RHA1
        for nRTPos in list_nRTPos:

            if sStrand == '+':
                nRTStart = nPAM_Nick  # PamNick -> 3'
                nRTEnd = nRTPos
                sRTSeq = sForTempSeq[nRTStart:nRTEnd]

            else:
                if self.sAltKey.startswith('sub'):
                    nRTStart = nRTPos
                    nRTEnd = nPAM_Nick  # PamNick -> 3'
                elif self.sAltKey.startswith('ins'):
                    nRTStart = nRTPos
                    nRTEnd = nPAM_Nick + self.nAltLen  # PamNick -> 3'
                elif self.sAltKey.startswith('del'):
                    nRTStart = nRTPos
                    nRTEnd = nPAM_Nick - self.nAltLen  # PamNick -> 3'

                sRTSeq = reverse_complement(sForTempSeq[nRTStart:nRTEnd])

                if not sRTSeq: continue
            # if END: sStrand

            sKey = len(sRTSeq)

            ## Set RT Length ##
            if nSetRTLen:
                if sKey != nSetRTLen: continue

            ## Limit Max RT len ##
            if sKey > nMaxRT: continue

            ## min RT from nick site to mutation ##
            if self.sAltKey.startswith('sub'):
                if sStrand == '+':
                    if sKey < abs(nAltIndex - nPAM_Nick): continue
                else:
                    if sKey < abs(nAltIndex - nPAM_Nick + self.nAltLen - 1): continue ### 
            else:
                if sStrand == '-':
                    if sKey < abs(nAltIndex - nPAM_Nick + self.nAltLen - 1): continue

            if self.sAltKey.startswith('ins'):
                if sKey < nAltPosWin + 1: continue

            if sKey not in dict_sRT:
                dict_sRT[sKey] = ''
            dict_sRT[sKey] = sRTSeq
        # loop END: nRTPos

        return [dict_sRT, dict_sPBS]


    # def END: determine_PBS_RT_seq

    def make_rt_pbs_combinations(self):
        for sPAMKey in self.dict_sSeqs:

            dict_sRT, dict_sPBS = self.dict_sSeqs[sPAMKey]

            list_sRT = [dict_sRT[sKey] for sKey in dict_sRT]
            list_sPBS = [dict_sPBS[sKey] for sKey in dict_sPBS]

            if sPAMKey not in self.dict_sCombos:
                self.dict_sCombos[sPAMKey] = ''
            self.dict_sCombos[sPAMKey] = {'%s,%s' % (sRT, sPBS): {} for sRT in list_sRT for sPBS in list_sPBS}
        # loop END: sPAMKey


    # def END: make_rt_pbs_combinations


    def determine_seqs(self):
        for sPAMKey in self.dict_sSeqs:

            sAltKey, sAltNotation, sStrand, nPAM_Nick, nAltPosWin, sPAMSeq, sGuideSeq = sPAMKey.split(',')
            nAltPosWin = int(nAltPosWin)
            nNickIndex = int(nPAM_Nick)

            for sSeqKey in self.dict_sCombos[sPAMKey]:

                sRTSeq, sPBSSeq = sSeqKey.split(',')

                ## for Tm1
                sForTm1 = reverse_complement_rna(sPBSSeq)

                if sStrand == '+':
                    ## for Tm2
                    sForTm2 = self.sWTSeq[nNickIndex:nNickIndex + len(sRTSeq)]
                    
                    ## for Tm3
                    if self.sAltType.startswith('sub'):
                        sForTm3 = self.sWTSeq[nNickIndex:nNickIndex + len(sRTSeq)]
                    elif self.sAltType.startswith('ins'):
                        sForTm3 = self.sWTSeq[nNickIndex:nNickIndex + len(sRTSeq) - self.nAltLen]
                    else:  # del
                        sForTm3 = self.sWTSeq[nNickIndex:nNickIndex + len(sRTSeq) + self.nAltLen]

                    ## for Tm4
                    if self.sAltType.startswith('sub'):
                        sTm4antiSeq = reverse_complement(self.sWTSeq[nNickIndex:nNickIndex + len(sRTSeq)])
                    elif self.sAltType.startswith('ins'):
                        sTm4antiSeq = reverse_complement(self.sWTSeq[nNickIndex:nNickIndex + len(sRTSeq) - self.nAltLen])
                    else:  # del
                        sTm4antiSeq = reverse_complement(self.sWTSeq[nNickIndex:nNickIndex + len(sRTSeq) + self.nAltLen])                    

                else:
                    ## for Tm2
                    sForTm2 = reverse_complement(self.sWTSeq[nNickIndex - len(sRTSeq):nNickIndex])

                    ## for Tm3
                    if self.sAltType.startswith('sub'):
                        sForTm3 = reverse_complement(self.sWTSeq[nNickIndex - len(sRTSeq):nNickIndex])
                    elif self.sAltType.startswith('ins'):
                        sForTm3 = reverse_complement(self.sWTSeq[nNickIndex - len(sRTSeq) + self.nAltLen:nNickIndex])
                    else:  # del
                        sForTm3 = reverse_complement(self.sWTSeq[nNickIndex - len(sRTSeq) - self.nAltLen:nNickIndex])

                    ## for Tm4
                    if self.sAltType.startswith('sub'):
                        sTm4antiSeq = self.sWTSeq[nNickIndex - len(sRTSeq):nNickIndex]
                    elif self.sAltType.startswith('ins'):
                        sTm4antiSeq = self.sWTSeq[nNickIndex - len(sRTSeq) + self.nAltLen:nNickIndex]
                    else:  # del
                        sTm4antiSeq = self.sWTSeq[nNickIndex - len(sRTSeq) - self.nAltLen:nNickIndex]

                # if END

                sForTm4 = [sRTSeq, sTm4antiSeq]

                ## for Tm5
                # sForTm5 = [reverse_complement(sRTSeq.replace('A', 'U')), sRTSeq] # original code
                sForTm5 = reverse_complement_rna(sRTSeq)

                self.dict_sCombos[sPAMKey][sSeqKey] = {
                    'Tm1_PBS': sForTm1,
                    'Tm2_RTT_cTarget_sameLength': sForTm2,
                    'Tm3_RTT_cTarget_replaced': sForTm3,
                    'Tm4_cDNA_PAM-oppositeTarget': sForTm4,
                    'Tm5_RTT_cDNA': sForTm5
                }
            # loop END: sSeqKey
        # loop END: sPAMKey
    # def END: determine_seqs


    def determine_secondary_structure(self):
        for sPAMKey in self.dict_sSeqs:

            sAltKey, sAltNotation, sStrand, nPAM_Nick, nAltPosWin, sPAMSeq, sGuideSeq = sPAMKey.split(',')
            list_sOutputKeys = ['Tm1_PBS', 'Tm2_RTT_cTarget_sameLength', 'Tm3_RTT_cTarget_replaced', 
                                'Tm4_cDNA_PAM-oppositeTarget', 'Tm5_RTT_cDNA', 'deltaTm_Tm4-Tm2', 'GC_count_PBS', 'GC_count_RTT', 'GC_count_RT-PBS',
                                'GC_contents_PBS', 'GC_contents_RTT', 'GC_contents_RT-PBS', 'MFE_RT-PBS-polyT', 'MFE_Spacer']

            if sPAMKey not in self.dict_sOutput:
                self.dict_sOutput[sPAMKey] = {}

            sGuideGN19 = 'G' + sGuideSeq[1:-3] 

            for sSeqKey in self.dict_sCombos[sPAMKey]:

                if sSeqKey not in self.dict_sOutput[sPAMKey]:
                    self.dict_sOutput[sPAMKey][sSeqKey] = {sKey: '' for sKey in list_sOutputKeys}

                self.determine_Tm(sPAMKey, sSeqKey)
                self.determine_GC(sPAMKey, sSeqKey)
                self.determine_MFE(sPAMKey, sSeqKey, sGuideGN19)

            # loop END: sSeqKey

        # loop END: sPAMKey


    def determine_Tm(self, sPAMKey, sSeqKey):

        sequences = [
            self.dict_sCombos[sPAMKey][sSeqKey]['Tm1_PBS'],
            self.dict_sCombos[sPAMKey][sSeqKey]['Tm2_RTT_cTarget_sameLength'],
            self.dict_sCombos[sPAMKey][sSeqKey]['Tm3_RTT_cTarget_replaced'],
            self.dict_sCombos[sPAMKey][sSeqKey]['Tm4_cDNA_PAM-oppositeTarget'],
            self.dict_sCombos[sPAMKey][sSeqKey]['Tm5_RTT_cDNA'],
        ]
        
        ## Tm1 DNA/RNA mm1 ##
        fTm1 = mt.Tm_NN(seq=Seq(sequences[0]), nn_table=mt.R_DNA_NN1)

        ## Tm2 DNA/DNA mm0 ##
        fTm2 = mt.Tm_NN(seq=Seq(sequences[1]), nn_table=mt.DNA_NN3)

        ## Tm3 DNA/DNA mm0 ##
        fTm3 = mt.Tm_NN(seq=Seq(sequences[2]), nn_table=mt.DNA_NN3)

        ## Tm4 DNA/DNA mm1 ##
        if not sequences[3]: fTm4 = 0

        else:
            for sSeq1, sSeq2 in zip(sequences[3][0], sequences[3][1]):
                try: fTm4 = mt.Tm_NN(seq=sSeq1, c_seq=sSeq2, nn_table=mt.DNA_NN3)
                except ValueError: continue
            # loop END: sSeq1, sSeq2
        # if END:

        # Tm5 - revcom(AAGTcGATCC(RNA version)) + AAGTcGATCC
        fTm5 = mt.Tm_NN(seq=Seq(sequences[4]), nn_table=mt.R_DNA_NN1)

        # deltaTm (Tm3 - Tm2)
        delta_tm = fTm3 - fTm2

        # 결과 저장
        self.dict_sOutput[sPAMKey][sSeqKey].update({
            'Tm1_PBS'                    : fTm1,
            'Tm2_RTT_cTarget_sameLength' : fTm2,
            'Tm3_RTT_cTarget_replaced'   : fTm3,
            'Tm4_cDNA_PAM-oppositeTarget': fTm4,
            'Tm5_RTT_cDNA'               : fTm5,
            'deltaTm_Tm4-Tm2'            : delta_tm,
        })

    # def END: determine_Tm


    def determine_GC(self, sPAMKey, sSeqKey):
        sRTSeqAlt, sPBSSeq = sSeqKey.split(',')

        self.dict_sOutput[sPAMKey][sSeqKey].update({
            'GC_count_PBS'      : sPBSSeq.count('G') + sPBSSeq.count('C'),
            'GC_count_RTT'      : sRTSeqAlt.count('G') + sRTSeqAlt.count('C'),
            'GC_count_RT-PBS'   : (sPBSSeq + sRTSeqAlt).count('G') + (sPBSSeq + sRTSeqAlt).count('C'),
            'GC_contents_PBS'   : 100 * gc(sPBSSeq),
            'GC_contents_RTT'   : 100 * gc(sRTSeqAlt),
            'GC_contents_RT-PBS': 100 * gc(sPBSSeq + sRTSeqAlt)
        })

    # def END: determine_GC

    def determine_MFE(self, sPAMKey, sSeqKey, sGuideGN19):

        sRTSeq, sPBSSeq = sSeqKey.split(',')

        sequences = [
            reverse_complement(sPBSSeq + sRTSeq) + 'TTTTTT', # RT + PBS + PolyT
            sGuideGN19                                       # GN19 guide seq
        ]

        # MFE 계산
        mfe_values = [self.calculate_MFE(seq) for seq in sequences]

        # 결과 저장
        self.dict_sOutput[sPAMKey][sSeqKey].update({
            'MFE_RT-PBS-polyT': mfe_values[0],
            'MFE_Spacer'      : mfe_values[1],
        })

    # def END: determine_MFE

    def calculate_MFE(self, sequence):
        # 시퀀스가 캐시에 있는지 확인
        if sequence in self.mfe_cache:
            return self.mfe_cache[sequence]

        # 시퀀스가 캐시에 없을 경우 계산
        mfe_value = round(fold_compound(sequence).mfe()[1], 1)
        self.mfe_cache[sequence] = mfe_value  # 계산된 값을 캐시에 저장
        return mfe_value
    

    def make_output_df(self):

        list_output = []
        list_sOutputKeys = ['Tm1_PBS', 'Tm2_RTT_cTarget_sameLength', 'Tm3_RTT_cTarget_replaced', 'Tm4_cDNA_PAM-oppositeTarget', 
                            'Tm5_RTT_cDNA', 'deltaTm_Tm4-Tm2', 'GC_count_PBS', 'GC_count_RTT', 'GC_count_RT-PBS',
                            'GC_contents_PBS', 'GC_contents_RTT', 'GC_contents_RT-PBS', 'MFE_RT-PBS-polyT', 'MFE_Spacer']

        for sPAMKey in self.dict_sSeqs:

            sAltKey, sAltNotation, sStrand, nPAM_Nick, nAltPosWin, sPAMSeq, sGuideSeq = sPAMKey.split(',')
            nNickIndex = int(nPAM_Nick)

            if sStrand == '+':
                sWTSeq74 = self.sWTSeq[nNickIndex - 21:nNickIndex + 53]
                nEditPos = 61 - nNickIndex
            else:
                sWTSeq74 = reverse_complement(self.sWTSeq[nNickIndex - 53:nNickIndex + 21])
                if not self.sAltType.startswith('ins'):
                    nEditPos = nNickIndex - 60 - self.nAltLen + 1
                else:
                    nEditPos = nNickIndex - 59

            for sSeqKey in self.dict_sOutput[sPAMKey]:

                dict_seq = self.dict_sCombos[sPAMKey][sSeqKey]
                sRTTSeq, sPBSSeq = sSeqKey.split(',')
                PBSlen = len(sPBSSeq)
                RTlen = len(sRTTSeq)

                sPBS_RTSeq = sPBSSeq + sRTTSeq
                s5Bufferlen = 21 - PBSlen
                s3Bufferlen = 53 - RTlen
                sEDSeq74 = 'x' * s5Bufferlen + sPBS_RTSeq + 'x' * s3Bufferlen

                if self.sAltType.startswith('del'):
                    RHA_len = len(sRTTSeq) - nEditPos + 1
                else:
                    RHA_len = len(sRTTSeq) - nEditPos - self.nAltLen + 1


                list_sOut = [self.input_id, reverse_complement(sPBS_RTSeq), len(sPBSSeq), len(sRTTSeq), len(sPBSSeq + sRTTSeq), 
                            nEditPos, self.nAltLen,RHA_len, sWTSeq74, sEDSeq74, 
                            self.type_sub, self.type_ins, self.type_del
                            ] + [self.dict_sOutput[sPAMKey][sSeqKey][sKey] for sKey in list_sOutputKeys]

                list_output.append(list_sOut)
            
            # loop END: sSeqKey
        
        hder_features = [
            # Sample ID
            'ID', 

            # pegRNA sequence features
            'RT-PBS', 

            # pegRNA edit and length features
            'PBS_len', 'RTT_len', 'RT-PBS_len', 'Edit_pos', 'Edit_len', 'RHA_len', 
            
            # Target sequences
            'Target', 'Masked_EditSeq', 

            # Edit types
            'type_sub', 'type_ins', 'type_del',

            # Tm features
            'Tm1_PBS', 'Tm2_RTT_cTarget_sameLength', 'Tm3_RTT_cTarget_replaced', 
            'Tm4_cDNA_PAM-oppositeTarget', 'Tm5_RTT_cDNA', 'deltaTm_Tm4-Tm2',

            # GC counts and contents
            'GC_count_PBS', 'GC_count_RTT', 'GC_count_RT-PBS', 
            'GC_contents_PBS', 'GC_contents_RTT', 'GC_contents_RT-PBS', 
            
            # RNA 2ndary structure features
            'MFE_RT-PBS-polyT', 'MFE_Spacer'
            ]

        df_out = pd.DataFrame(list_output, columns=hder_features)

        list_spacer = [wt74[24-self.spacer_len:24] for wt74 in df_out.Target]
        df_out.insert(1, 'Spacer', list_spacer)
        
        # loop END: sPAMKey

        return df_out

# def END: make_output


class GeneInteractionModel(nn.Module):


    def __init__(self, hidden_size, num_layers, num_features=24, dropout=0.1):
        super(GeneInteractionModel, self).__init__()
        self.hidden_size = hidden_size
        self.num_layers = num_layers

        self.c1 = nn.Sequential(
            nn.Conv2d(in_channels=4, out_channels=128, kernel_size=(2, 3), stride=1, padding=(0, 1)),
            nn.BatchNorm2d(128),
            nn.GELU(),
        )
        self.c2 = nn.Sequential(
            nn.Conv1d(in_channels=128, out_channels=108, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm1d(108),
            nn.GELU(),
            nn.AvgPool1d(kernel_size=2, stride=2),

            nn.Conv1d(in_channels=108, out_channels=108, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm1d(108),
            nn.GELU(),
            nn.AvgPool1d(kernel_size=2, stride=2),

            nn.Conv1d(in_channels=108, out_channels=128, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm1d(128),
            nn.GELU(),
            nn.AvgPool1d(kernel_size=2, stride=2),
        )

        self.r = nn.GRU(128, hidden_size, num_layers, batch_first=True, bidirectional=True)

        self.s = nn.Linear(2 * hidden_size, 12, bias=False)

        self.d = nn.Sequential(
            nn.Linear(num_features, 96, bias=False),
            nn.ReLU(),
            nn.Dropout(dropout),
            nn.Linear(96, 64, bias=False),
            nn.ReLU(),
            nn.Dropout(dropout),
            nn.Linear(64, 128, bias=False)
        )

        self.head = nn.Sequential(
            nn.BatchNorm1d(140),
            nn.Dropout(dropout),
            nn.Linear(140, 1, bias=True),
        )

    def forward(self, g, x):
        g = torch.squeeze(self.c1(g), 2)
        g = self.c2(g)
        g, _ = self.r(torch.transpose(g, 1, 2))
        g = self.s(g[:, -1, :])

        x = self.d(x) ## 여기가 문제
        out = self.head(torch.cat((g, x), dim=1))
        return F.softplus(out)

def seq_concat(data, col1='Target', col2='Masked_EditSeq', seq_length=74):
    wt = preprocess_seq(data[col1], seq_length)
    # ed = preprocess_masked_seq(data[col2], seq_length)
    ed = preprocess_seq(data[col2], seq_length)
    g = np.concatenate((wt, ed), axis=1)
    g = 2 * g - 1

    return g


def select_cols(data):
    features = data.loc[:, ['PBS_len', 'RTT_len', 'RT-PBS_len', 'Edit_pos', 'Edit_len', 'RHA_len', 'type_sub',
                            'type_ins', 'type_del', 'Tm1_PBS', 'Tm2_RTT_cTarget_sameLength', 'Tm3_RTT_cTarget_replaced', 
                            'Tm4_cDNA_PAM-oppositeTarget', 'Tm5_RTT_cDNA', 'deltaTm_Tm4-Tm2',
                            'GC_count_PBS', 'GC_count_RTT', 'GC_count_RT-PBS', 
                            'GC_contents_PBS', 'GC_contents_RTT', 'GC_contents_RT-PBS', 
                            'MFE_RT-PBS-polyT', 'MFE_Spacer', 'DeepSpCas9_score']]

    return features




import os, requests, inspect
from genet import models
from tqdm import tqdm

'''
Model finder repo
Get your models from here!
'''

class LoadModel:
    def __init__(self, model:str, effector:str, cell_type:str=None):
        """Locates the path and file of the deep learning model loaded in the predict module. 
        If the model has not been installed yet, it downloads the necessary model files from the genet-models GitHub repository.

        Args:
            model (str): The type of model to be loaded. Possible options include 'DeepSpCas9', 'DeepCas9variants', 'DeepSmallCas9', and 'DeepPrime'.
            effector (str): The specific effector types available for each model.
            cell_type (str, optional): The specific cell type options available for each model. Currently, this additional selection is only available for DeepPrime. Defaults to None.

        Raises:
            KeyError: An error occurs if you attempt to load a model that is not provided by GenET.
        """        

        if model == 'DeepSpCas9':
            model_type = effector
        elif model == 'DeepCas9variants':
            model_type = effector
        elif model == 'DeepSmallCas9':
            model_type = effector
        elif model == 'DeepPrime':
            model_type = effector + '-' + cell_type
        else: 
            model_type = effector + '-' + cell_type
        
        # Check if this model is supported by GenET
        try   : self.info = models.constants.dict_model_info[model_type]
        except: raise KeyError('Not available model in GenET!')
        
        self.model_dir = inspect.getfile(models).replace('__init__.py', '') + self.info['path']

        # Download the model if it has not been downloaded yet
        if not os.path.exists(self.model_dir):
            os.makedirs(self.model_dir)

            dict_files = models.constants.dict_model_requests

            self.download_from_github(
                repo     = self.info['repo'],
                path     = self.info['path'],
                files    = dict_files[self.info['type']],
                save_dir = self.model_dir,
                )
    
    # def End: __init__

    def download_from_github(self, repo:str, path:str, files:str, save_dir:str):
        """A function to download files from a specified path in a GitHub repository. 
        It uses `requests.get` to download from the GitHub URL.

        Args:
            repo (str): The GitHub repository where the file to be downloaded is located.
            path (str): The path within the repository where the file is located.
            files (str): The name of the file to download.
            save_dir (str): The location to download the file.
        """        
        
        print('The model %s is not installed. Download checkpoint files.' % path)

        for file_name in files:

            url = f"https://raw.githubusercontent.com/{repo}/{path}/{file_name}"
            save_path = f"{save_dir}/{file_name}"

            response = requests.get(url, stream=True)
            response.raise_for_status()
            total_size = int(response.headers.get('content-length', 0))
            
            with open(save_path, "wb") as file:
                
                for data in tqdm(response.iter_content(chunk_size=4096), total=total_size//4096, 
                                 unit="KB", desc="Downloading", unit_scale=True, leave=False):
                    
                    file.write(data)
            
            print(f"[Info] File downloaded: {save_path}")

    # def End: download_from_github




# from genet.utils import __version__

'''
genet의 처음 init 부분
'''

# How to release with 2FA accounts

# [주의!] 먼저 아래의 version을 변경한다!!
__version__ = '0.17.2'

# hatch build
# twine upload --username __token__ dist/*
# [Enter PyPI API token]



import sys
import numpy as np
from genet.analysis.constants import BLOSUM62
from collections import namedtuple


__all__ = ["AlignmentResult", "aligner"]


IS_PY2 = False
if sys.version_info.major == 2:
    IS_PY2 = True

# Container for alignment result
AlignmentResult = namedtuple(
    'AlignmentResult',
    ['seq1', 'seq2', 'start1', 'start2',
     'end1', 'end2', 'n_gaps1', 'n_gaps2',
     'n_mismatches', 'score'])


def aligner(seqj:str, seqi:str, method:str='global', 
            gap_open:float=-7, gap_extend:float=-7, gap_double:float=-7,
            matrix:dict=BLOSUM62, max_hits:int=1):
    '''Calculates the alignment of two sequences.

    The supported 'methods' are:
        * 'global' for a global Needleman-Wunsh algorithm
        * 'local' for a local Smith-Waterman alignment
        * 'global_cfe' for a global alignment with cost-free ends
        * 'glocal' for an alignment which is 'global' only with respect to
          the shorter sequence (also known as a 'semi-global' alignment)

    Returns the aligned (sub)sequences as character arrays.

    Gotoh, O. (1982). J. Mol. Biol. 162, 705-708.
    Needleman, S. & Wunsch, C. (1970). J. Mol. Biol. 48(3), 443-53.
    Smith, T.F. & Waterman M.S. (1981). J. Mol. Biol. 147, 195-197.

    Arguments:

        - seqj First aligned iterable object of symbols.
        - seqi Second aligned iterable object of symbols.
        - method Type of alignment: 'global', 'global_cfe', 'local', 'glocal'.
        - gap_open The gap-opening cost.
        - gap_extend The cost of extending an open gap.
        - gap_double The gap-opening cost if a gap is already open
          in the other sequence.
        - matrix  A score matrix dictionary.
        - max_hits The maximum number of results to return in
          case multiple alignments with the same score are found. If set to 1,
          a single ``AlignmentResult`` object is returned. If set to values
          larger than 1, a list containing ``AlignmentResult`` objects are
          returned. If set to `None`, all alignments with the maximum score
          are returned.
    '''
    assert max_hits is None or max_hits > 0
    NONE, LEFT, UP, DIAG = range(4)  # NONE is 0
    GAP_CHAR = ord('-') if not IS_PY2 else '-'
    max_j = len(seqj)
    max_i = len(seqi)

    if max_j > max_i:
        flip = 1
        seqi, seqj = seqj, seqi
        max_i, max_j = max_j, max_i
    else:
        flip = 0

    seqi = seqi.encode() if not isinstance(seqi, bytes) else seqi
    seqj = seqj.encode() if not isinstance(seqj, bytes) else seqj

    F = np.zeros((max_i + 1, max_j + 1), dtype=np.float32)
    I = np.ndarray((max_i + 1, max_j + 1), dtype=np.float32)
    I.fill(-np.inf)
    J = np.ndarray((max_i + 1, max_j + 1), dtype=np.float32)
    J.fill(-np.inf)
    pointer = np.zeros((max_i + 1, max_j + 1), dtype=np.uint)  # NONE

    if method == 'global':
        pointer[0, 1:] = LEFT
        pointer[1:, 0] = UP
        F[0, 1:] = gap_open + gap_extend * \
            np.arange(0, max_j, dtype=np.float32)
        F[1:, 0] = gap_open + gap_extend * \
            np.arange(0, max_i, dtype=np.float32)
    elif method == 'global_cfe':
        pointer[0, 1:] = LEFT
        pointer[1:, 0] = UP
    elif method == 'glocal':
        pointer[0, 1:] = LEFT
        F[0, 1:] = gap_open + gap_extend * \
            np.arange(0, max_j, dtype=np.float32)

    for i in range(1, max_i + 1):
        ci = seqi[i - 1:i]
        for j in range(1, max_j + 1):
            cj = seqj[j - 1:j]
            # I
            I[i, j] = max(
                         F[i, j - 1] + gap_open,
                         I[i, j - 1] + gap_extend,
                         J[i, j - 1] + gap_double)
            # J
            J[i, j] = max(
                         F[i - 1, j] + gap_open,
                         J[i - 1, j] + gap_extend,
                         I[i - 1, j] + gap_double)
            # F
            diag_score = F[i - 1, j - 1] + matrix[cj][ci]
            left_score = I[i, j]
            up_score = J[i, j]
            max_score = max(diag_score, up_score, left_score)

            F[i, j] = max(0, max_score) if method == 'local' else max_score

            if method == 'local':
                if F[i, j] == 0:
                    pass  # point[i,j] = NONE
                elif max_score == diag_score:
                    pointer[i, j] = DIAG
                elif max_score == up_score:
                    pointer[i, j] = UP
                elif max_score == left_score:
                    pointer[i, j] = LEFT
            elif method == 'glocal':
                # In a semi-global alignment we want to consume as much as
                # possible of the longer sequence.
                if max_score == up_score:
                    pointer[i, j] = UP
                elif max_score == diag_score:
                    pointer[i, j] = DIAG
                elif max_score == left_score:
                    pointer[i, j] = LEFT
            else:
                # global
                if max_score == up_score:
                    pointer[i, j] = UP
                elif max_score == left_score:
                    pointer[i, j] = LEFT
                else:
                    pointer[i, j] = DIAG

    # container for traceback coordinates
    ij_pairs = []
    if method == 'local':
        # max anywhere
        maxv_indices = np.argwhere(F == F.max())[:max_hits]
        for index in maxv_indices:
            ij_pairs.append(index)
    elif method == 'glocal':
        # max in last col
        max_score = F[:, -1].max()
        maxi_indices = np.argwhere(F[:, -1] == F[:, -1].max())\
            .flatten()[:max_hits]
        for i in maxi_indices:
            ij_pairs.append((i, max_j))
    elif method == 'global_cfe':
        # from i,j to max(max(last row), max(last col)) for free
        row_max = F[-1].max()
        col_max = F[:, -1].max()
        # expecting max to exist on either last column or last row
        if row_max > col_max:
            col_idces = np.argwhere(F[-1] == row_max).flatten()[:max_hits]
            for cid in col_idces:
                ij_pairs.append((i, cid))
        elif row_max < col_max:
            row_idces = np.argwhere(F[:, -1] == col_max).flatten()[:max_hits]
            for rid in row_idces:
                ij_pairs.append((rid, j))
        # special case: max is on last row, last col
        elif row_max == col_max == F[i, j]:
            # check if max score also exist on other cells in last row
            # or last col. we expect only one of the case.
            col_idces = np.argwhere(F[-1] == row_max).flatten()
            row_idces = np.argwhere(F[:, -1] == col_max).flatten()
            ncol_idces = len(col_idces)
            nrow_idces = len(row_idces)

            # tiebreaker between row/col is whichever has more max scores
            if ncol_idces > nrow_idces:
                for cid in col_idces[:max_hits]:
                    ij_pairs.append((i, cid))
            elif ncol_idces < nrow_idces:
                for rid in row_idces[:max_hits]:
                    ij_pairs.append((rid, j))
            elif ncol_idces == nrow_idces == 1:
                ij_pairs.append((i, j))
            else:
                raise RuntimeError('Unexpected multiple maximum global_cfe'
                                   ' scores.')
        else:
            raise RuntimeError('Unexpected global_cfe scenario.')
    else:
        # method must be global at this point
        ij_pairs.append((i, j))

    results = []
    for i, j in ij_pairs:
        align_j = []
        align_i = []
        score = F[i, j]
        p = pointer[i, j]
        # mimic Python's coord system
        if method.startswith("global"):
            end_i, end_j = max_i, max_j
        else:
            end_i, end_j = i, j
        n_gaps_i, n_gaps_j, n_mmatch = 0, 0, 0

        # special case for global_cfe ~ one cell may contain multiple pointer
        # directions
        if method == 'global_cfe':
            if i < max_i:
                align_i.extend([c for c in seqi[i:][::-1]])
                align_j.extend([GAP_CHAR] * (max_i - i))
                n_gaps_j += 1
            elif j < max_j:
                align_i.extend([GAP_CHAR] * (max_j - j))
                align_j.extend([c for c in seqj[j:][::-1]])
                n_gaps_i += 1

        while p != NONE:
            if p == DIAG:
                i -= 1
                j -= 1
                ichar = seqi[i]
                jchar = seqj[j]
                if ichar != jchar:
                    n_mmatch += 1
                align_j.append(jchar)
                align_i.append(ichar)
            elif p == LEFT:
                j -= 1
                align_j.append(seqj[j])
                if not align_i or align_i[-1] != GAP_CHAR:
                    n_gaps_i += 1
                align_i.append(GAP_CHAR)
            elif p == UP:
                i -= 1
                align_i.append(seqi[i])
                if not align_j or align_j[-1] != GAP_CHAR:
                    n_gaps_j += 1
                align_j.append(GAP_CHAR)
            else:
                raise Exception('wtf!')
            p = pointer[i, j]

        align_i = bytes(align_i[::-1]) \
            if not IS_PY2 else ''.join(align_i[::-1])
        align_j = bytes(align_j[::-1]) \
            if not IS_PY2 else ''.join(align_j[::-1])

        aln = (AlignmentResult(align_i, align_j, i, j, end_i, end_j,
                               n_gaps_i, n_gaps_j, n_mmatch, score)
               if flip else
               AlignmentResult(align_j, align_i, j, i, end_j, end_i,
                               n_gaps_j, n_gaps_i, n_mmatch, score))

        results.append(aln)

    return results


import tensorflow as tf
import numpy as np
import pandas as pd

from genet.predict.PredUtils import *
from genet.models import LoadModel

import os
from Readfile import *
from Process import *
import random
random.seed(123)

from absl import app
from absl import flags

import datetime
##############################################################################

FLAGS = flags.FLAGS
flags.DEFINE_enum('mode', None, ['Sa-KKH', 'Sauri', 'Sauri-KKH', 'Cj','Nm1','Nm2','Sa','St1'], 'mode of training')
flags.DEFINE_string('filename', None, 'test filename')
flags.DEFINE_integer('fold', 5, 'number of folds')

class SmallCas:
    def __init__(self, effector:str, gpu_env=0):
        '''DeepSmallCas9 score function  
        
        The list_target30 should have a 30bp sequence in the form of a list.
        '''
        self.effector = effector

        #TensorFlow config
        conf = tf.compat.v1.ConfigProto()
        conf.gpu_options.allow_growth = True
        os.environ['CUDA_VISIBLE_DEVICES'] = str(gpu_env)

        model = LoadModel('DeepSmallCas9', effector)

        self.length    = model.model_info['length']
        self.t_length  = model.model_info['t_length']
        self.bio_num   = model.model_info['bio_num']
        self.model_tag = model.model_info['model_tag']
        
        model_dir  = model.model_dir
        model_save = '%s/%s' % (model_dir, self.model_tag)



    def predict(self, list_target: list) -> pd.DataFrame:
        '''Input으로 30nt target context sequence 들이 담긴 list가 들어오면,
        각 sequence 마다의 prediction score를 계산해서 list로 return 하는 함수
        '''
        
        TEST_X = preprocess_seq(list_target, self.length)
        TEST_mod_X = [test_dict['onehot_mod_seq']]
        TEST_bio = [test_dict['bio']]
        TEST_Y = [test_dict['val']]
        test_row = 0
        test_col = 0




##########################################


        # Initiate Xlsx Output Files
        testbook = xlsxwriter.Workbook(modulename+"/TEST_OUTPUT.xlsx")
        testsheet = [testbook.add_worksheet('{}'.format(test_param['fname'][:-5][:31]))] # -5 is for erasing .xlsx
        test_dict = getfile(data_path, test_param, length, t_length)
        TEST_X = [test_dict['onehot_seq']]
        TEST_mod_X = [test_dict['onehot_mod_seq']]
        TEST_bio = [test_dict['bio']]
        TEST_Y = [test_dict['val']]
        test_row = 0
        test_col = 0

        for seq, mod_seq in zip(test_dict['seq'], test_dict['mod_seq']):
            testsheet[-1].write(test_row, test_col, seq)
            testsheet[-1].write(test_row, test_col+1, mod_seq)
            test_row += 1

        best_model_path_list = [FLAGS.mode+'/Model/{}/best_model'.format(os.listdir(FLAGS.mode+'/Model')[-1])]
        best_model_list                     = []

        for best_model_path in best_model_path_list:
            for best_modelname in os.listdir(best_model_path):
                if "meta" in best_modelname:
                    best_model_list.append(best_modelname[:-5])

        print(best_model_list)

        best_model_path = best_model_path_list[0]
        best_model      = best_model_list[0]
        valuelist       = best_model.split('-')
        fulllist        = []

        for value in valuelist:
            try:
                value=int(value)
            except:
                try:    value=float(value)
                except: pass
            fulllist.append(value)

        print(fulllist[2:])

        filter_size_1, filter_size_2, filter_size_3, filter_num_1, filter_num_2, filter_num_3, l_rate, load_episode, node_1, node_2 = fulllist[2:]
        filter_size = [filter_size_1, filter_size_2, filter_size_3]
        filter_num  = [filter_num_1, filter_num_2, filter_num_3]

        args = [filter_size, filter_num, l_rate, 0, None, node_1, node_2]
        # Loading the model with the best validation score and test
        tf.compat.v1.reset_default_graph()
        with tf.compat.v1.Session(config=conf) as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            model = Model.Model(filter_size, filter_num, length, t_length, node_1, node_2, args[2], bio_num)
            saver = tf.compat.v1.train.Saver()
            saver.restore(sess, best_model_path+"/PreTrain-Final-{}-{}-{}-{}-{}-{}-{}-{}-{}-{}".format(args[0][0], args[0][1], args[0][2], args[1][0], args[1][1], args[1][2], args[2], load_episode, args[5], args[6]))
            Model_test(sess, TEST_X[0], model, args, load_episode, testbook, testsheet[0], col_index=test_col+2, modulename=modulename, TEST_mod_X=TEST_mod_X[0], TEST_bio=TEST_bio[0], TEST_Y=TEST_Y[0])
            testbook.close()




def main(_):
    data_path = 'dataset/'
    # Sequence, Window, Frequency(Proportion), Wt or Alt, File Name, Sheet Name
    test_param  = {}
    test_param['init'] = 1
    test_param['sheet'] = 'Sheet1'
    modelname = "Model"
    test_param['seq'] = 0
    test_param['mod_seq'] = 1
    test_param['val'] = 2

    if FLAGS.mode in ['Cj']:
        length = 22
        t_length = 4+22+8+3
        bio_num = 275
        modename = "CjCas9"
        
    elif FLAGS.mode in ['efSa']:
        length = 21
        t_length = 34
        bio_num = 263
        modename = "efSaCas9"

    elif FLAGS.mode in ['enCj']:
        length = 22
        t_length = 37
        bio_num = 275
        modename = "enCjCas9"
        
    elif FLAGS.mode in ['eSa']:
        length = 21
        t_length = 34
        bio_num = 263
        modename = "eSaCas9"

    elif FLAGS.mode in ['Nm1']:
        length = 23
        t_length = 4+23+8+3
        bio_num = 287
        modename = "Nm1Cas9"

    elif FLAGS.mode in ['Nm2']:
        length = 23
        t_length = 4+23+7+3
        bio_num = 287
        modename = "Nm2Cas9"

    elif FLAGS.mode in ['Sa']:
        length = 21
        t_length = 4+21+6+3
        bio_num = 263
        modename = "SaCas9"
        
    elif FLAGS.mode in ['Sa-HF']:
        length = 21
        t_length = 34
        bio_num = 263
        modename = "SaCas9-HF"

    elif FLAGS.mode in ['Sa-KKH']:
        length = 21
        t_length = 4+21+6+3
        bio_num = 263
        modename = "SaCas9-KKH"

    elif FLAGS.mode in ['Sa-KKH-HF']:
        length = 21
        t_length = 34
        bio_num = 263
        modename = "SaCas9-KKH-HF"
        
    elif FLAGS.mode in ['Sa-Slug']:
        length = 21
        t_length = 32
        bio_num = 263
        modename = "Sa-SlugCas9"
        
    elif FLAGS.mode in ['Sauri']:
        length = 21
        t_length = 4+21+4+3
        bio_num = 263
        modename = "SauriCas9"

    elif FLAGS.mode in ['Sauri-KKH']:
        length = 21
        t_length = 4+21+4+3
        bio_num = 263
        modename = "SauriCas9-KKH"
        
    elif FLAGS.mode in ['Slug']:
        length = 21
        t_length = 32
        bio_num = 263
        modename = "SlugCas9"
        
    elif FLAGS.mode in ['Slug-HF']:
        length = 21
        t_length = 32
        bio_num = 263
        modename = "SlugCas9-HF"
        
    elif FLAGS.mode in ['sRGN']:
        length = 21
        t_length = 32
        bio_num = 263
        modename = "sRGN3.1"
        
    elif FLAGS.mode in ['St1']:
        length = 19
        t_length = 4+19+6+3
        bio_num = 239
        modename = "St1Cas9"

    else:
        raise NotImplementedError

    test_param['bio']   = test_param['val'] + bio_num
    if FLAGS.filename is None:
        test_param['fname'] = FLAGS.mode+"_sample.xlsx"
    else:
        test_param['fname'] = FLAGS.filename # "FILENAME"

    total_fold = FLAGS.fold

    #TensorFlow config
    conf = tf.compat.v1.ConfigProto()
    conf.gpu_options.allow_growth = True
    os.environ['CUDA_VISIBLE_DEVICES'] = '0'

    Model = __import__(modelname)
    modulename = FLAGS.mode + "_Test/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S") + "/"

    if not os.path.exists(modulename):
        os.makedirs(modulename)

    # Initiate Xlsx Output Files
    testbook = xlsxwriter.Workbook(modulename+"/TEST_OUTPUT.xlsx")
    testsheet = [testbook.add_worksheet('{}'.format(test_param['fname'][:-5][:31]))] # -5 is for erasing .xlsx
    test_dict = getfile(data_path, test_param, length, t_length)
    TEST_X = [test_dict['onehot_seq']]
    TEST_mod_X = [test_dict['onehot_mod_seq']]
    TEST_bio = [test_dict['bio']]
    TEST_Y = [test_dict['val']]
    test_row = 0
    test_col = 0

    for seq, mod_seq in zip(test_dict['seq'], test_dict['mod_seq']):
        testsheet[-1].write(test_row, test_col, seq)
        testsheet[-1].write(test_row, test_col+1, mod_seq)
        test_row += 1

    best_model_path_list = [FLAGS.mode+'/Model/{}/best_model'.format(os.listdir(FLAGS.mode+'/Model')[-1])]
    best_model_list                     = []

    for best_model_path in best_model_path_list:
        for best_modelname in os.listdir(best_model_path):
            if "meta" in best_modelname:
                best_model_list.append(best_modelname[:-5])

    print(best_model_list)

    best_model_path = best_model_path_list[0]
    best_model      = best_model_list[0]
    valuelist       = best_model.split('-')
    fulllist        = []

    for value in valuelist:
        try:
            value=int(value)
        except:
            try:    value=float(value)
            except: pass
        fulllist.append(value)

    print(fulllist[2:])

    filter_size_1, filter_size_2, filter_size_3, filter_num_1, filter_num_2, filter_num_3, l_rate, load_episode, node_1, node_2 = fulllist[2:]
    filter_size = [filter_size_1, filter_size_2, filter_size_3]
    filter_num  = [filter_num_1, filter_num_2, filter_num_3]

    args = [filter_size, filter_num, l_rate, 0, None, node_1, node_2]
    # Loading the model with the best validation score and test
    tf.compat.v1.reset_default_graph()
    with tf.compat.v1.Session(config=conf) as sess:
        sess.run(tf.compat.v1.global_variables_initializer())
        model = Model.Model(filter_size, filter_num, length, t_length, node_1, node_2, args[2], bio_num)
        saver = tf.compat.v1.train.Saver()
        saver.restore(sess, best_model_path+"/PreTrain-Final-{}-{}-{}-{}-{}-{}-{}-{}-{}-{}".format(args[0][0], args[0][1], args[0][2], args[1][0], args[1][1], args[1][2], args[2], load_episode, args[5], args[6]))
        Model_test(sess, TEST_X[0], model, args, load_episode, testbook, testsheet[0], col_index=test_col+2, modulename=modulename, TEST_mod_X=TEST_mod_X[0], TEST_bio=TEST_bio[0], TEST_Y=TEST_Y[0])
        testbook.close()


def create_new_conv_layer(input_data, num_input_channels, num_filters, filter_shape, stride_shape, name, is_training):
    conv_filt_shape = [filter_shape[0], filter_shape[1], num_input_channels,
                        num_filters]
    weights   = tf.compat.v1.Variable(tf.compat.v1.truncated_normal(conv_filt_shape, stddev=0.03),
                                        name=name+'_W')
    bias      = tf.compat.v1.Variable(tf.compat.v1.truncated_normal([num_filters]), name=name+'_b')

    out_layer = tf.compat.v1.nn.conv2d(input_data, weights, [1, stride_shape[0], stride_shape[1], 1], padding='VALID')
    out_layer += bias
    out_layer = tf.compat.v1.layers.dropout(tf.compat.v1.nn.relu(out_layer), 0.3, is_training)
    return out_layer

class Model(object):
    def __init__(self, filter_size, filter_num, length, t_length, node_1 = 80, node_2 = 60, l_rate = 0.005, bio_num=0):
        self.inputs         = tf.compat.v1.placeholder(tf.compat.v1.float32, [None, 1, length, 4])
        self.mod_inputs     = tf.compat.v1.placeholder(tf.compat.v1.float32, [None, 1, t_length, 4])
        self.bios           = tf.compat.v1.placeholder(tf.compat.v1.float32, [None, bio_num])
        self.targets        = tf.compat.v1.placeholder(tf.compat.v1.float32, [None, 1])
        self.is_training    = tf.compat.v1.placeholder(tf.compat.v1.bool)

        L_filter_num = 4
        stride = 1
        if filter_num[0] == 0:
            raise NotImplementedError
        else:
            L_pool_0 = create_new_conv_layer(self.inputs, L_filter_num, filter_num[0]*3, [1, filter_size[0]], [1, stride], name='conv1', is_training=self.is_training)
            L_pool_1 = create_new_conv_layer(self.mod_inputs, L_filter_num, filter_num[0]*3, [1, filter_size[0]], [1, stride], name='conv2', is_training=self.is_training)

        with tf.compat.v1.variable_scope('Fully_Connected_Layer1'):
            layer_node_0 = int((length-filter_size[0])/stride)+1
            node_num_0   = layer_node_0*filter_num[0]*3
            L_flatten_0  = tf.compat.v1.reshape(L_pool_0, [-1, node_num_0])
            layer_node_1 = int((t_length-filter_size[0])/stride)+1
            node_num_1   = layer_node_1*filter_num[0]*3
            L_flatten_1  = tf.compat.v1.reshape(L_pool_1, [-1, node_num_1])
            
            L_flatten_concat = tf.compat.v1.concat([L_flatten_0, L_flatten_1, self.bios], 1, name='concat')

            W_fcl1       = tf.compat.v1.get_variable("W_fcl1", shape=[node_num_0+node_num_1+bio_num, node_1])
            B_fcl1       = tf.compat.v1.get_variable("B_fcl1", shape=[node_1])
            L_fcl1_pre   = tf.compat.v1.nn.bias_add(tf.compat.v1.matmul(L_flatten_concat, W_fcl1), B_fcl1)
            L_fcl1       = tf.compat.v1.nn.relu(L_fcl1_pre)
            L_fcl1_drop  = tf.compat.v1.layers.dropout(L_fcl1, 0.3, self.is_training)

        with tf.compat.v1.variable_scope('Fully_Connected_Layer2'):
            W_fcl2       = tf.compat.v1.get_variable("W_fcl2", shape=[node_1, node_2])
            B_fcl2       = tf.compat.v1.get_variable("B_fcl2", shape=[node_2])
            L_fcl2_pre   = tf.compat.v1.nn.bias_add(tf.compat.v1.matmul(L_fcl1_drop, W_fcl2), B_fcl2)
            L_fcl2       = tf.compat.v1.nn.relu(L_fcl2_pre)
            L_fcl2_drop  = tf.compat.v1.layers.dropout(L_fcl2, 0.3, self.is_training)

        with tf.compat.v1.variable_scope('Output_Layer'):
            W_out        = tf.compat.v1.get_variable("W_out", shape=[node_2, 1])
            B_out        = tf.compat.v1.get_variable("B_out", shape=[1])
            self.outputs = tf.compat.v1.nn.bias_add(tf.compat.v1.matmul(L_fcl2_drop, W_out), B_out)

        # Define loss function and optimizer
        self.obj_loss    = tf.compat.v1.reduce_mean(tf.compat.v1.square(self.targets - self.outputs))
        optimizer = tf.compat.v1.train.AdamOptimizer(learning_rate=l_rate)
        self.gvs = optimizer.compute_gradients(self.obj_loss)
        capped_gvs = [(tf.compat.v1.clip_by_value(grad, -1., 1.), var) for grad, var in self.gvs]
        self.train_op = optimizer.apply_gradients(capped_gvs)

if __name__ == '__main__':
  app.run(main)


import genet
import genet.utils
from genet.predict.PredUtils import *


def pe_score(Ref_seq: str, 
            ED_seq: str, 
            sAlt: str,
            sID:str       = 'Sample',
            pe_system:str = 'PE2max',
            cell_type:str = 'HEK293T',
            pbs_min:int   = 7,
            pbs_max:int   = 15,
            rtt_max:int   = 40,
            show_features = False,
            silence:bool  = False,
            ):
    '''### genet.predict.pe_score is deprecated. 
    Please consider genet.predict.DeepPrime instead.'''

    raise DeprecationWarning('''genet.predict.pe_score is deprecated. Please consider genet.predict.DeepPrime instead.''')
    

def pecv_score(cv_record,
               sID:str       = 'Sample',
               pe_system:str = 'PE2max',
               cell_type:str = 'HEK293T',
               pbs_min:int   = 7,
               pbs_max:int   = 15,
               rtt_max:int   = 40
               ):

    '''### genet.predict.pecv_score is deprecated.'''

    raise DeprecationWarning('''genet.predict.pecv_score is deprecated.''')
    



'''
각종 database에 대한 config가 정리되어 있는 파일.
예를 들어, URL / database 구조 / 접속시 일반적으로 들어가는 input 정보들 등을 정리한다.
genet.database.config에서 

'''

import genet
import inspect, os
from datetime import datetime

class DBconfig:
    def __init__(self) -> None:

        self.genet_path = inspect.getfile(genet).replace('__init__.py', '')

    def get_file_version(self, file_path):

        # 파일의 생성일자 및 수정일자 가져오기
        created_time  = os.path.getctime(file_path)

        # 생성일자와 수정일자를 날짜 및 시간 형식으로 변환
        created_date  = datetime.fromtimestamp(created_time).strftime('%Y-%m-%d')

        return created_date
  



class Ensemblconfig(DBconfig):
    def __init__(self) -> None:
        super().__init__()
        pass






import genet
import os, sys, regex, glob, shutil
import pandas as pd
import multiprocessing as mp
from Bio import SeqIO
from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor, as_completed
from genet.utils import run_FLASH, split_fq_file


'''
TODO
1. flash를 python code로 구현한 것이 없으므로, 여기서 input은 .fastq 파일만 가능
2. 나중에 flashpy 개발이 좀 더 진행되면 도입하는 것을 생각해보자
3. python 기본적으로 python 3.7~3.10까지 호환되는 것을 목표로 하고, 3.11도 테스트하기

'''


def reverse_complement(sSeq):
    dict_sBases = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N', '.': '.', '*': '*',
                   'a': 't', 'c': 'g', 'g': 'c', 't': 'a'}
    list_sSeq   = list(sSeq)  # Turns the sequence in to a gigantic list
    list_sSeq   = [dict_sBases[sBase] for sBase in list_sSeq]
    return ''.join(list_sSeq)[::-1]
#def END: reverse_complement

class cPEData:
    def __init__(self):
        
        pass


def load_PE_input (sInFile):
    dict_sOutput = {}
    with open(sInFile, 'r') as InFile:
        for sReadLine in InFile:
            ## File Format ##
            ## Target#  | Barcode | WT_Target | Edit_Target
            ## 181      | TTT.... | CTGCC..   | CTGCC...

            if sReadLine.endswith('edit\n'): continue ## SKIP HEADER
            if sReadLine.endswith('#\n'): continue ## SKIP HEADER
            if sReadLine.startswith('#'): continue ## SKIP HEADER

            if ',' in sReadLine:
                list_sColumn = sReadLine.strip('\n').split(',')
            else:
                list_sColumn = sReadLine.strip('\n').split('\t')

            cPE           = cPEData()
            cPE.sBarcode  = list_sColumn[0].upper()
            cPE.sRefSeq   = list_sColumn[1].upper()
            cPE.sWTSeq    = list_sColumn[2].upper()
            cPE.sAltSeq   = list_sColumn[3].upper()

            if len(list_sColumn) == 6:
                cPE.sIntendedOnly = list_sColumn[4].upper()
                cPE.sMisMatchOnly = list_sColumn[5].upper()

            barcode = cPE.sBarcode

            if barcode not in dict_sOutput:
                dict_sOutput[barcode] = ''
            dict_sOutput[barcode] = cPE

    return dict_sOutput
#def END: load_PE_input

def load_read_data (dict_sOutFreq, dict_sOutRead, sInFile):

    with open(sInFile, 'r') as InFile:
        for sReadLine in InFile:

            list_sColumn = sReadLine.strip('\n').split('\t')
            sBarcode     = list_sColumn[0]

            list_WT      = [sReadID for sReadID in list_sColumn[1].split(',') if sReadID]
            list_edited  = [sReadID for sReadID in list_sColumn[2].split(',') if sReadID]
            list_other   = [sReadID for sReadID in list_sColumn[3].split(',') if sReadID]

            dict_sOutFreq[sBarcode][0] += len(list_WT)
            dict_sOutFreq[sBarcode][1] += len(list_edited)
            dict_sOutFreq[sBarcode][2] += len(list_other)

            ### Collect sequence information ###
            # dict_sOutRead[sBarcode][0] += list_WT
            # dict_sOutRead[sBarcode][1] += list_edited
            # dict_sOutRead[sBarcode][2] += list_other


class TargetAnalyzer:
    def __init__(self,):
        # flash 있는지 확인
        self.flash = '/extdata1/GS/flash'

        pass

    def setup(self, sAnalysis:str, file1:str, file2:str, barcode_file:str, sRE:str='[T]{4}[ACGT]{14}', min_overlap:int=5, overwrite:bool=False):
        
        ## 지금 돌릴 분석 (run)의 결과를 담을 경로 세팅
        self.sAnalysis = sAnalysis
        self.Out_DIR   = f'{self.sAnalysis}_results'
        self.Tmp_DIR   = f'{self.Out_DIR}/tmp'
        self.Log_File  = f'{self.Out_DIR}/log.log'
        os.makedirs(self.Tmp_DIR, exist_ok=True)

        ## Options
        self.sRE  = sRE
        
        ## Load Barcode Data -> Dict에서 DataFrame으로 변경하기
        self.dict_cPE = load_PE_input(barcode_file)

        ## Step 1: Run FLASH; Paired-end NGS read combining
        flash_log = run_FLASH(self.flash, file1, file2, self.Tmp_DIR, out_prefix=self.sAnalysis, min_overlap=min_overlap, overwrite=overwrite)
        with open(self.Log_File, 'a') as outfile:
            outfile.write(flash_log)

        ## Step 2: Split FASTQ files
        sFastqTag = f'{self.sAnalysis}.extendedFrags'
        self.list_sSplitFile = split_fq_file(self.Tmp_DIR, sFastqTag, overwrite)


    def run(self, error_type:str='SwitchingFree', nRefBuffer=24, nCores:int=30):
        # nRefBuffer설정: SF인 경우, 앞에서부터 spacer 위치까지, EF인 경우, pegRNA 전체 길이까지
        # Barcode list를 보고, SF인 경우에 자동으로 RefBuffer를 설정하는 코드를 추가하기

        ## 설정 1: Error Type: BarcodeOnly, ErrorFree (error가 있는 pegRNA를 filtering 하는 여부에 대한 설정)
        ## BarcodeOnly = pegRNA (spacer+scaffold+RTPBS+polyT) 부분에 error나 mismatch가 있어도 barcode counting 하겠다.
        ## ErrorFree   = pegRNA (spacer+scaffold+RTPBS+polyT) 부분이 우리가 디자인한대로 완벽히 일치하는 경우만 barcode counting.
        if error_type not in ['SwitchingFree', 'ErrorFree', 'BarcodeOnly']:
            raise KeyError(f'Not available key. Please select SwitchingFree, ErrorFree, or BarcodeOnly')

        ## Run Analysis ##
        self._barcode_sorter (error_type, nRefBuffer, nCores)
        self._combine_output_pickle (error_type)


    def _barcode_sorter(self, error_type, nRefBuffer, nCores):
    
        # Create parameters list
        list_sParameters = []
        
        for sSplitFile in self.list_sSplitFile:
            
            split_dir = f'{self.Tmp_DIR}/split'
            sInFastq  = f'{split_dir}/{sSplitFile}'
            sSplitTag = sSplitFile.replace('.fq', '').split('extendedFrags_')[1] # sSplitTag 예시: fastq_0001

            list_sParameters.append({
                'split_dir'  : split_dir,
                'split_tag'  : sSplitTag,
                'split_fq'   : sInFastq,
                'barcode'    : self.dict_cPE,
                're_pattern' : self.sRE,
                'error_type' : error_type,
                'nRefBuffer' : nRefBuffer,
            })

        # Use ProcessPoolExecutor for parallel processing
        with ProcessPoolExecutor(max_workers=nCores) as executor:
            # Submit tasks for `worker` and track progress
            list_work = [executor.submit(self._worker, params) for params in list_sParameters]
            for _ in tqdm(as_completed(list_work), total=len(list_work), desc="Sorting by Barcode", ncols=100, ascii=' ='):
                pass  # Wait for all worker tasks to complete


    def _worker (self, params):

        split_dir  = params['split_dir']
        sSplitTag  = params['split_tag']
        sInFastq   = params['split_fq']
        dict_cPE   = params['barcode']
        sRE        = params['re_pattern']
        sError     = params['error_type']
        nRefBuffer = params['nRefBuffer']

        dict_sBarcodes = {}
        
        if sError == 'ErrorFree':
            nRefBuffer = -nRefBuffer  # Barcode length to subtract from back of RefSeq
        elif sError == 'SwitchingFree':
            nRefBuffer = nRefBuffer

        # Process input FASTQ file
        # 이 부분을 나중에 gzip 파일로 작업하는 것으로 변경하기? 
        # Tmp file의 크기를 줄일 수 있는 방법을 고민해보자
        with open(sInFastq, 'r') as InFile:
            for sSeqData in SeqIO.parse(InFile, 'fastq'):

                sReadID = str(sSeqData.id)
                sNGSSeq = str(sSeqData.seq)

                for sReIndex in regex.finditer(sRE, sNGSSeq, overlapped=True):
                    nIndexStart   = sReIndex.start()
                    nIndexEnd     = sReIndex.end()
                    sBarcodeMatch = sNGSSeq[nIndexStart:nIndexEnd] # sRGN
                    sRefSeqCheck  = sNGSSeq[:nIndexStart]
                    
                    ### Skip Non-barcodes ###
                    try: cPE = dict_cPE[sBarcodeMatch]
                    except KeyError:continue

                    ### Skip error in Refseq ###
                    if sError != 'ErrorProne':
                        if cPE.sRefSeq[:nRefBuffer] not in sRefSeqCheck:
                            continue

                    if sBarcodeMatch not in dict_sBarcodes:
                        dict_sBarcodes[sBarcodeMatch] = []
                    dict_sBarcodes[sBarcodeMatch].append([sReadID, sNGSSeq, nIndexEnd])

        # Generate output dictionary
        dict_sOutput = {sBarcode: {'WT': [], 'Alt': [], 'Other': []} for sBarcode in dict_sBarcodes}
        for sBarcode, read_info in dict_sBarcodes.items():
            cPE = dict_cPE[sBarcode]

            for sReadID, sNGSSeq, sIndexS in read_info:
                sTargetCheck  = reverse_complement(sNGSSeq[sIndexS-1:])

                if cPE.sWTSeq in sTargetCheck:
                    dict_sOutput[sBarcode]['WT'].append(cPE.sWTSeq)
                elif cPE.sAltSeq in sTargetCheck:
                    dict_sOutput[sBarcode]['Alt'].append(cPE.sAltSeq)
                else:
                    dict_sOutput[sBarcode]['Other'].append(cPE.sAltSeq)
                
        # Write output to file
        sOutFile   = f'{split_dir}/{sSplitTag}.reads.txt'
        list_Types = ['WT', 'Alt', 'Other']

        with open(sOutFile, 'w') as OutFile:
            for sBarcode in dict_sOutput:
                sOut = '%s\t%s\n' % (sBarcode, '\t'.join([','.join(dict_sOutput[sBarcode][sAltType]) for sAltType in list_Types]))
                OutFile.write(sOut)

    #def END: worker
    

    def _combine_output_pickle (self, error_type):

        nSplitNo   = 4
        nFileCnt   = len(self.list_sSplitFile)
        list_nBins = [[int(nFileCnt * (i + 0) / nSplitNo), int(nFileCnt * (i + 1) / nSplitNo)] for i in range(nSplitNo)]

        list_sKeys     = ['WT', 'Alt', 'Other',]
        dict_sOutFreq  = {sBarcode: [0 for i in list_sKeys] for sBarcode in self.dict_cPE}
        
        for nStart, nEnd in tqdm(list_nBins, total=len(list_nBins), desc='Make temp output', ncols=100, ascii=' ='):
            list_sSubSplit  = self.list_sSplitFile[nStart:nEnd]
            dict_sOutFreq_Tmp = {sBarcode: [0, 0, 0,] for sBarcode in self.dict_cPE}
            dict_sOutRead_Tmp = {sBarcode: [[], [], [],] for sBarcode in self.dict_cPE}

            for sSplitFile in list_sSubSplit:
                sSplitTag = sSplitFile.split('extendedFrags_')[1].replace('.fq', '') # sSplitTag 예시: fastq_0001
                sInFile   = f'{self.Tmp_DIR}/split/{sSplitTag}.reads.txt'
                
                assert os.path.isfile(sInFile)
                load_read_data (dict_sOutFreq_Tmp, dict_sOutRead_Tmp, sInFile)

            for sBarcode in self.dict_cPE:

                dict_sOutFreq[sBarcode][0] += dict_sOutFreq_Tmp[sBarcode][0]
                dict_sOutFreq[sBarcode][1] += dict_sOutFreq_Tmp[sBarcode][1]
                dict_sOutFreq[sBarcode][2] += dict_sOutFreq_Tmp[sBarcode][2]

        sHeader  = '%s\t%s\t%s\n' % ('Barcode', '\t'.join(list_sKeys), 'Total')
        sOutFile = f'{self.Out_DIR}/{self.sAnalysis}_{error_type}.combinedFreq.txt'
        
        with open(sOutFile, 'w') as OutFile:
            OutFile.write(sHeader)
            
            for sBarcode in self.dict_cPE:

                list_sOut = [str(sOutput) for sOutput in dict_sOutFreq[sBarcode]]
                nTotal  = sum(dict_sOutFreq[sBarcode])
                sOut    = '%s\t%s\t%s\n' % (sBarcode, '\t'.join(list_sOut), nTotal)
                OutFile.write(sOut)
                
    #def END: combine_output
#def END: TargetAnalyzer


class SortByBarcodes:
    '''# SortByBarcodes

    This class makes new fastq files only containing specific barcode sequences.
    The barcode list should be input files as DNA sequences.
    The barcode pattern is string based on regular expressions.
    Now, only fastq format is available for input data file.

    #### Example
    >>> from genet import analysis as ans
    >>> ans.SortByBarcodes('./MyNGS.fastq', './Barcode_pattern.csv', 'TCGTATGCCGTCTTCTGCTTG[ATGC]{14}', n_cores=10)

    The output file will be generated in current working directory in default.
    If you want to save your output at other path, you can set the 'output_path' option.

    #### Example
    >>> ans.SortByBarcodes(seq_file='./MyNGS.fastq',
                           barcode_file='./Barcode_pattern.csv',
                           barcode_pattern='TCGTATGCCGTCTTCTGCTTG[ATGC]{14}',
                           output_name='My_sorted_data',
                           output_path='/extdata/mydir/results',
                           n_cores=20
                           )
    '''
    
    def __init__(self,
                 seq_file:str,
                 barcode_file:str,
                 barcode_pattern:str = None,
                 output_name:str = 'barcode_sorted', 
                 output_path:str = './',
                 data_format:str = 'fastq',
                 output_format:str = 'fastq',
                 n_cores:int = int(mp.cpu_count()*0.5),
                 remove_temp_files:bool = True,
                 silence:bool = False,
                 ):

        # check input types
        if n_cores > mp.cpu_count():
            sys.exit('n_core should be lower than the number of cores which your machine has')

        # load barcode and data files
        self.df_bc    = pd.read_csv(barcode_file, names=['id', 'barcode'])

        splits = genet.utils.SplitFastq(
            seq_file, n_cores, out_path=output_path, out_name=output_name, silence=silence)
        
        self.list_sParameters = []
        for s in splits.names:
            self.list_sParameters.append([self.df_bc, barcode_pattern, s, splits.dir, output_format, silence])

        # sorting barcodes from subsplit files
        p = mp.Pool(n_cores)
        if silence == False: print('[Info] Starting map_async: sorting by barcodes')
        p.map_async(self._sort_barcode, self.list_sParameters).get()

        p.close()
        p.join()

        # combine all temp files
        if silence == False: print('[Info] Make final sorted files')
        sOUT_DIR = '%s/%s_sorted' % (output_path, output_name)
        os.makedirs(sOUT_DIR, exist_ok=True)

        self.couts = mp.Manager().dict()
        self.couts['_Not_matched'] = []
        for key in self.df_bc['barcode']: self.couts[key] = []

        self.barcodes = self.couts.keys()
        self.nBarcodeCnt = len(self.barcodes)

        list_barcode_nBins = [[int(self.nBarcodeCnt * (i + 0) / n_cores), int(self.nBarcodeCnt * (i + 1) / n_cores)] for i in range(n_cores)]

        if silence == False: print('[Info] Make barcode subsplits')
        self.list_combine_param = [[splits.dir, output_format, sOUT_DIR, self.couts, self.barcodes[nStart:nEnd], silence] for nStart, nEnd in list_barcode_nBins]

        p = mp.Pool(n_cores)
        if silence == False: print('[Info] Starting map_async: combine all temp files')
        p.map_async(self._combine_files, self.list_combine_param).get()

        p.close()
        p.join()

        # finalize
        if remove_temp_files==True:
            if silence == False: print('[Info] Removing temp files')
            shutil.rmtree(splits.dir)
        if silence==False: print('[Info] Done: SortByBarcodes - %s' % output_name)

    # def END: __init__

    def summary(self):
        '''### Summary of the barcode sorting results
        After sorting NGS reads by barcodes, the average read counts or distributions per barcode is shown by this method.

        #### Example
        >>> from genet import analysis as ans
        >>> sorting = ans.SortByBarcodes('./MyNGS.fastq', './Barcode_pattern.csv', 'TCGTATGCCGTCTTCTGCTTG[ATGC]{14}', n_cores=10)
        >>> sorting.summary()
        '''

        print('')

    
    def rankplot(self):
        '''### Rank plot of read count distribution
        
        >>> from genet import analysis as ans
        >>> sorting = ans.SortByBarcodes('./MyNGS.fastq', './Barcode_pattern.csv', 'TCGTATGCCGTCTTCTGCTTG[ATGC]{14}', n_cores=10)
        >>> sorting.rankplot()
        
        '''
# class END: SortByBarcodes


    def _sort_barcode(self, list_sParameters):
        '''Sorting the fastq file by barcode list
        '''

        # parameters
        df_bc            = list_sParameters[0]
        barcode_pattern  = list_sParameters[1]
        subsplit_name    = list_sParameters[2]
        subsplit_dir     = list_sParameters[3]
        output_format    = list_sParameters[4]
        silence          = list_sParameters[5]
        
        if silence == False: print('[Info] Barcode sorting - %s' % (subsplit_name))
        
        # make temp dir
        temp_dir = '%s/_temp_%s_sorting' % (subsplit_dir, subsplit_name)
        os.makedirs(temp_dir, exist_ok=True)

        dict_barcode = {'_Not_matched': []}
        for key in df_bc['barcode']:
            dict_barcode[key] = []
        
        fq_file = '%s/%s' % (subsplit_dir, subsplit_name)
        record_iter = SeqIO.parse(open(fq_file), output_format)

        for rec in record_iter:
            seq = str(rec.seq)
            check_match = False
            
            if barcode_pattern == None:
                for k in dict_barcode.keys():
                    if k not in seq: continue
                    else:
                        dict_barcode[k].append(rec)
                        check_match=True
                        break

            else:
                try:
                    for sReIndex in regex.finditer(barcode_pattern, seq, overlapped=True):
                        nIndexStart = sReIndex.start()
                        nIndexEnd = sReIndex.end()
                        window = seq[nIndexStart:nIndexEnd]
                        
                        try: 
                            dict_barcode[window].append(rec)
                            check_match = True
                            break
                        except KeyError: continue
                except KeyError: continue
            
            if check_match==False: dict_barcode['_Not_matched'].append(rec)
                
        if silence == False: print('Make temp sorted %s file: %s' % (output_format, subsplit_name))
        
        for barcode, seq_rec in dict_barcode.items():
            SeqIO.write(seq_rec, '%s/%s.%s' % (temp_dir, barcode, output_format), output_format)

    # def END: sort_barcode
 
    def _combine_files(self, list_combine_param):
        """Combine files by name"""

        # parameters
        splits_dir    = list_combine_param[0]
        output_format = list_combine_param[1]
        sOUT_DIR      = list_combine_param[2]
        counts        = list_combine_param[3]
        barcodes      = list_combine_param[4]
        silence       = list_combine_param[5]

        for key in barcodes:
            if silence == False: print('Make combined file: %s' % key)
            temp_fqs = glob.glob('%s/**/%s.%s' % (splits_dir, key, output_format))

            output_file_name = '%s/%s.%s' % (sOUT_DIR, key, output_format)

            with open(output_file_name, 'w') as outfile:
                for filename in sorted(temp_fqs):
                    with open(filename) as file:        
                        outfile.write(file.read())
            
            with open(output_file_name, 'r') as outfile:
                counts[key] = len(outfile.readlines())


def sort_barcode_and_combine(list_sParameters):
    '''Sorting the fastq file by barcode list
    sorting 하는 동시에 바로 with open으로 최종 파일에 기록하는 방법도 가능할까?
    그럼 별도의 temp 파일을 만들지 않아도 되기 때문에 훨씬 시간도 빠르고 IO 소모가 없을듯.
    일단 작은 사이즈로 구현해보고 검증해봐야 함. 
    '''

    # parameters
    df_bc            = list_sParameters[0]
    barcode_pattern  = list_sParameters[1]
    subsplit_name    = list_sParameters[2]
    subsplit_dir     = list_sParameters[3]
    output_format    = list_sParameters[4]
    silence          = list_sParameters[5]
    
    if silence == False: print('[Info] Barcode sorting - %s' % (subsplit_name))
    
    # make temp dir
    temp_dir = '%s/_temp_%s_sorting' % (subsplit_dir, subsplit_name)
    os.makedirs(temp_dir, exist_ok=True)

    dict_barcode = {'_Not_matched': []}
    for key in df_bc['barcode']:
        dict_barcode[key] = []
    
    fq_file = '%s/%s' % (subsplit_dir, subsplit_name)
    record_iter = SeqIO.parse(open(fq_file), output_format)

    for rec in record_iter:
        seq = str(rec.seq)
        check_match = False
        
        if barcode_pattern == None:
            for k in dict_barcode.keys():
                if k not in seq: continue
                else:
                    dict_barcode[k].append(rec)
                    check_match=True
                    break

        else:
            try:
                for sReIndex in regex.finditer(barcode_pattern, seq, overlapped=True):
                    nIndexStart = sReIndex.start()
                    nIndexEnd = sReIndex.end()
                    window = seq[nIndexStart:nIndexEnd]
                    
                    try: 
                        dict_barcode[window].append(rec)
                        check_match = True
                        break
                    except KeyError: continue
            except KeyError: continue
        
        if check_match==False: dict_barcode['_Not_matched'].append(rec)
            
    if silence == False: print('Make temp sorted %s file: %s' % (output_format, subsplit_name))
    
    for barcode, seq_rec in dict_barcode.items():
        SeqIO.write(seq_rec, '%s/%s.%s' % (temp_dir, barcode, output_format), output_format) 

""" Codon usage analysis "temporary" """

def calculate_codon_composition(seq):
    """Calculates the frequency of each codon in a DNA sequence."""
    codon_counts = {}
    for i in range(0, len(seq) - 2, 3):
        codon = seq[i:i+3]
        codon_counts[codon] = codon_counts.get(codon, 0) + 1
 
    total_count = sum(codon_counts.values())
    for codon, count in codon_counts.items():
        codon_counts[codon] = count / total_count

    return codon_counts

def find_orfs(seq):
    """Identifies potential open reading frames (ORFs) in a DNA sequence."""
    orfs = []
    for frame in range(3):
        for start in range(frame, len(seq), 3):
            codon = seq[start:start + 3]
            if codon == 'ATG':
                end = start + 3
                while end < len(seq) and seq[end:end + 3] not in ['TAA', 'TAG', 'TGA']:
                    end += 3
                orfs.append((start, end, '+'))
    return orfs




import pandas as pd

'''
GG PAM 에서 mutation이 생겨서 PAM disruption이 일어날 때,
prime editing efficiency가 증가하는 순위
Reference: Yu et al. Cell 2023
'''

dict_pam_disrup_rank = {
    'AT': 1,
    'TA': 2,
    'AG': 3,
    'TT': 4,
    'TG': 5,
    'CT': 6,
    'TC': 7,
    'AC': 8,
    'GT': 9,
    'CC': 10,
    'AA': 11,
    'CA': 12,
    'GA': 13,
    'GC': 14,
    'CG': 15,
    'GG': 16,
}


def test_score_data(example_id:str) -> dict:
    """DeepPrime output 예시 파일을 불러오는 함수

    Args:
        example_id (str): Example 파일의 번호 (example_1, example_2, example_3, example_4)

    Returns:
        dict: _description_
    """    

    dict_data = {
        'example_1': {
            'file'      : './tests/test_data/ABL1_ex4_pos26C_A_output.parquet',
            'test_idx'  : 9,
            'ref_seq'   : 'CTGTCTCTGTGGGCTGAAGGCTGTTCCCTGTTTCCTTCAGCTCTACGTCTCCTCCGAGAGCCGCTTCAACACCCTGGCCGAGTTGGTTCATCATCATTCAACGGTGGCCGACGGGCTCATC',
            'frame'     : 2,
            'cds_start' : 39,
            'cds_end'   : 121,
        },

        'example_2': {
            'file'      : './tests/test_data/ABL1_ex4_pos27C_A_output.parquet',
            'test_idx'  : 9,
            'ref_seq'   : 'TGTCTCTGTGGGCTGAAGGCTGTTCCCTGTTTCCTTCAGCTCTACGTCTCCTCCGAGAGCCGCTTCAACACCCTGGCCGAGTTGGTTCATCATCATTCAACGGTGGCCGACGGGCTCATCA',
            'frame'     : 0,
            'cds_start' : 39,
            'cds_end'   : 121,
        },


        'example_3': {
            'file'      : './tests/test_data/ABL1_ex4_pos189A_G_output.parquet',
            'test_idx'  : 587,
            'ref_seq'   : 'GTCTATGGTGTGTCCCCCAACTACGACAAGTGGGAGATGGAACGCACGGACATCACCATGAAGCACAAGCTGGGCGGGGGCCAGTACGGGGAGGTGTACGAGGGCGTGTGGAAGAAATACA',
            'frame'     : 0,
            'cds_start' : 0,
            'cds_end'   : 121,
        },


        'example_4': {
            'file'      : './tests/test_data/ABL1_ex4_pos155C_A_output.parquet',
            'test_idx'  : 148,
            'ref_seq'   : 'CCATTATCCAGCCCCAAAGCGCAACAAGCCCACTGTCTATGGTGTGTCCCCCAACTACGACAAGTGGGAGATGGAACGCACGGACATCACCATGAAGCACAAGCTGGGCGGGGGCCAGTAC',
            'frame'     : 2,
            'cds_start' : 0,
            'cds_end'   : 121,
        },
        
        'example_5': {
            'file'      : './tests/test_data/ABL1_ex4_pos40T_C_output.parquet',
            'test_idx'  : 148,
            'ref_seq'   : 'CCATTATCCAGCCCCAAAGCGCAACAAGCCCACTGTCTATGGTGTGTCCCCCAACTACGACAAGTGGGAGATGGAACGCACGGACATCACCATGAAGCACAAGCTGGGCGGGGGCCAGTAC',
            'frame'     : 2,
            'cds_start' : 0,
            'cds_end'   : 121,
        },

        'example_6': {
            'file'      : './tests/test_data/ABL1_ex4_pos123C_A_output.parquet',
            'test_idx'  : 148,
            'ref_seq'   : 'tcaacggtggccgacgggctcatcaccacgctccattatccagccccaaagcgcaacaagcccactgtctatggtgtgtcccccaactacgacaagtgggagatggaacgcacggacatca',
            'frame'     : 0,
            'cds_start' : 0,
            'cds_end'   : 121,
        },
    }

    selected = dict_data[example_id]

    return pd.read_parquet(selected['file']).iloc[selected['test_idx']], dict_data[example_id]

